# Copyright Notice:
# Copyright 2018-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

import os
import time
from collections import Counter, OrderedDict, defaultdict
from collections import namedtuple
from bs4 import BeautifulSoup
from functools import lru_cache
import os.path

from redfish_service_validator.helper import getNamespace, getNamespaceUnversioned

import logging
my_logger = logging.getLogger(__name__)


EDM_NAMESPACE = "http://docs.oasis-open.org/odata/ns/edm"
EDMX_NAMESPACE = "http://docs.oasis-open.org/odata/ns/edmx"
EDM_TAGS = ['Action', 'Annotation', 'Collection', 'ComplexType', 'EntityContainer', 'EntityType', 'EnumType', 'Key',
            'Member', 'NavigationProperty', 'Parameter', 'Property', 'PropertyRef', 'PropertyValue', 'Record',
            'Schema', 'Singleton', 'Term', 'TypeDefinition']
EDMX_TAGS = ['DataServices', 'Edmx', 'Include', 'Reference']


def bad_edm_tags(tag):
    return tag.namespace == EDM_NAMESPACE and tag.name not in EDM_TAGS


def bad_edmx_tags(tag):
    return tag.namespace == EDMX_NAMESPACE and tag.name not in EDMX_TAGS


def other_ns_tags(tag):
    return tag.namespace != EDM_NAMESPACE and tag.namespace != EDMX_NAMESPACE


def reference_missing_uri_attr(tag):
    return tag.name == 'Reference' and tag.get('Uri') is None


def include_missing_namespace_attr(tag):
    return tag.name == 'Include' and tag.get('Namespace') is None


def format_tag_string(tag):
    tag_name = tag.name if tag.prefix is None else tag.prefix + ':' + tag.name
    tag_attr = ''
    for attr in tag.attrs:
        tag_attr += '{}="{}" '.format(attr, tag.attrs[attr])
    return (tag_name + ' ' + tag_attr).strip()


def list_html(entries):
    html_str = '<ul>'
    for entry in entries:
        html_str += '<li>{}</li>'.format(entry)
    html_str += '</ul>'
    return html_str


def tag_list_html(tags_dict):
    html_str = '<ul>'
    for tag in tags_dict:
        html_str += '<li>{} {}</li>' \
            .format(tag, '(' + str(tags_dict[tag]) + ' occurrences)' if tags_dict[tag] > 1 else '')
    html_str += '</ul>'
    return html_str


class Metadata(object):
    metadata_uri = '/redfish/v1/$metadata'
    schema_type = '$metadata'

    def __init__(self, data, service, logger):
        logger.info('Constructing metadata...')
        self.success_get = False
        self.service = service
        self.uri_to_namespaces = defaultdict(list)
        self.elapsed_secs = 0
        self.metadata_namespaces = set()
        self.service_namespaces = set()
        self.schema_store = dict()
        self.bad_tags = dict()
        self.bad_tag_ns = dict()
        self.refs_missing_uri = dict()
        self.includes_missing_ns = dict()
        self.bad_schema_uris = set()
        self.bad_namespace_include = set()
        self.counter = OrderedCounter()
        self.logger = logger
        self.redfish_extensions_alias_ok = False

        start = time.time()
        self.md_soup = None
        self.service_refs = None
        uri = Metadata.metadata_uri

        self.elapsed_secs = time.time() - start
        self.schema_obj = None
        if data:
            self.md_soup = BeautifulSoup(data, "xml")
            self.service_refs = getReferenceDetails(self.md_soup)
            self.success_get = True
            # set of namespaces included in $metadata
            self.metadata_namespaces = {k for k in self.service_refs.keys()}
            # create map of schema URIs to namespaces from $metadata
            for k in self.service_refs.keys():
                self.uri_to_namespaces[self.service_refs[k][1]].append(self.service_refs[k][0])
            logger.debug('Metadata: uri = {}'.format(uri))
            logger.debug('Metadata: metadata_namespaces: {} = {}'
                         .format(type(self.metadata_namespaces), self.metadata_namespaces))
            # check for Redfish alias for RedfishExtensions.v1_0_0
            ref = self.service_refs.get('Redfish')
            if ref is not None and ref[0] == 'RedfishExtensions.v1_0_0':
                self.redfish_extensions_alias_ok = True
            logger.debug('Metadata: redfish_extensions_alias_ok = {}'.format(self.redfish_extensions_alias_ok))
            # check for XML tag problems
            self.check_tags()
            # check that all namespace includes are found in the referenced schema
            self.check_namespaces_in_schemas()
            logger.debug('Metadata: bad_tags = {}'.format(self.bad_tags))
            logger.debug('Metadata: bad_tag_ns = {}'.format(self.bad_tag_ns))
            logger.debug('Metadata: refs_missing_uri = {}'.format(self.refs_missing_uri))
            logger.debug('Metadata: includes_missing_ns = {}'.format(self.includes_missing_ns))
            logger.debug('Metadata: bad_schema_uris = {}'.format(self.bad_schema_uris))
            logger.debug('Metadata: bad_namespace_include = {}'.format(self.bad_namespace_include))
            for ref in self.service_refs:
                name, uri = self.service_refs[ref]
                success, soup, origin = getSchemaDetails(service, name, uri)
                self.schema_store[name] = soup
        else:
            logger.warning('Metadata: getSchemaDetails() did not return success')

    def get_schema_obj(self):
        return self.schema_obj

    def get_soup(self):
        return self.md_soup

    def get_service_refs(self):
        return self.service_refs

    def get_metadata_namespaces(self):
        return self.metadata_namespaces

    def get_service_namespaces(self):
        return self.service_namespaces

    def add_service_namespace(self, namespace):
        self.service_namespaces.add(namespace)

    def get_missing_namespaces(self):
        return self.service_namespaces - self.metadata_namespaces

    def get_schema_uri(self, namespace):
        ref = self.service_refs.get(namespace)
        if ref is not None:
            return ref[1]
        else:
            return None

    def check_tags(self):
        """
        Perform some checks on the tags in the $metadata XML looking for unrecognized tags,
        tags missing required attributes, etc.
        """
        try:
            for tag in self.md_soup.find_all(bad_edm_tags):
                tag_str = format_tag_string(tag)
                self.bad_tags[tag_str] = self.bad_tags.get(tag_str, 0) + 1
            for tag in self.md_soup.find_all(bad_edmx_tags):
                tag_str = format_tag_string(tag)
                self.bad_tags[tag_str] = self.bad_tags.get(tag_str, 0) + 1
            for tag in self.md_soup.find_all(reference_missing_uri_attr):
                tag_str = format_tag_string(tag)
                self.refs_missing_uri[tag_str] = self.refs_missing_uri.get(tag_str, 0) + 1
            for tag in self.md_soup.find_all(include_missing_namespace_attr):
                tag_str = format_tag_string(tag)
                self.includes_missing_ns[tag_str] = self.includes_missing_ns.get(tag_str, 0) + 1
            for tag in self.md_soup.find_all(other_ns_tags):
                tag_str = tag.name if tag.prefix is None else tag.prefix + ':' + tag.name
                tag_ns = 'xmlns{}="{}"'.format(':' + tag.prefix if tag.prefix is not None else '', tag.namespace)
                tag_str = tag_str + ' ' + tag_ns
                self.bad_tag_ns[tag_str] = self.bad_tag_ns.get(tag_str, 0) + 1
        except Exception as e:
            self.logger.warning('Metadata: Problem parsing $metadata document: {}'.format(e))

    def check_namespaces_in_schemas(self):
        """
        Check that all namespaces included from a schema URI are actually in that schema
        """
        for k in self.uri_to_namespaces.keys():
            schema_uri = k
            if '#' in schema_uri:
                schema_uri, frag = k.split('#', 1)
            schema_type = os.path.basename(os.path.normpath(k)).strip('.xml').strip('_v1')
            success, soup, _ = getSchemaDetails(self.service, schema_type, schema_uri)
            if success:
                for namespace in self.uri_to_namespaces[k]:
                    if soup.find('Schema', attrs={'Namespace': namespace}) is None:
                        msg = 'Namespace {} not found in schema {}'.format(namespace, k)
                        self.logger.debug('Metadata: {}'.format(msg))
                        self.bad_namespace_include.add(msg)
            else:
                self.logger.error('Metadata: failure opening schema {} of type {}'.format(schema_uri, schema_type))
                self.bad_schema_uris.add(schema_uri)

    def get_counter(self):
        """
        Create a Counter instance containing the counts of any errors found
        """
        counter = OrderedCounter()
        # informational counters
        counter['metadataNamespaces'] = len(self.metadata_namespaces)
        counter['serviceNamespaces'] = len(self.service_namespaces)
        # error counters
        counter['missingRedfishAlias'] = 0 if self.redfish_extensions_alias_ok else 1
        counter['missingNamespaces'] = len(self.get_missing_namespaces())
        counter['badTags'] = len(self.bad_tags)
        counter['missingUriAttr'] = len(self.refs_missing_uri)
        counter['missingNamespaceAttr'] = len(self.includes_missing_ns)
        counter['badTagNamespaces'] = len(self.bad_tag_ns)
        counter['badSchemaUris'] = len(self.bad_schema_uris)
        counter['badNamespaceInclude'] = len(self.bad_namespace_include)
        self.counter = counter
        return self.counter

    def to_html(self):
        """
        Convert the $metadata validation results to HTML
        """
        time_str = 'response time {0:.6f}s'.format(self.elapsed_secs)
        section_title = '{} ({})'.format(Metadata.metadata_uri, time_str)

        counter = self.get_counter()

        html_str = ''
        html_str += '<tr><th class="titlerow bluebg"><b>{}</b></th></tr>'\
            .format(section_title)
        html_str += '<tr><td class="titlerow"><table class="titletable"><tr>'
        html_str += '<td class="title" style="width:40%"><div>{}</div>\
                        <div class="button warn" onClick="document.getElementById(\'resMetadata\').classList.toggle(\'resultsShow\');">Show results</div>\
                        </td>'.format(section_title)
        html_str += '<td class="titlesub log" style="width:30%"><div><b>Schema File:</b> {}</div><div><b>Resource Type:</b> {}</div></td>'\
            .format(Metadata.metadata_uri, Metadata.schema_type)
        html_str += '<td style="width:10%"' + \
            ('class="pass"> GET Success' if self.success_get else 'class="fail"> GET Failure') + '</td>'
        html_str += '<td style="width:10%">'

        errors_found = False
        for count_type in counter.keys():
            style = 'class=log'
            if 'bad' in count_type or 'missing' in count_type:
                if counter[count_type] > 0:
                    errors_found = True
                    style = 'class="fail log"'
            html_str += '<div {style}>{p}: {q}</div>'.format(
                    p=count_type, q=counter.get(count_type, 0), style=style)

        html_str += '</td></tr>'
        html_str += '</table></td></tr>'
        html_str += '<tr><td class="results" id=\'resMetadata\'><table><tr><th>$metadata validation results</th></tr>'

        if self.success_get and not errors_found:
            html_str += '<tr><td class="pass log">Validation successful</td></tr>'
        elif not self.success_get:
            html_str += '<tr><td class="fail log">ERROR - Unable to retrieve $metadata resource at {}</td></tr>'\
                .format(Metadata.metadata_uri)
        else:
            if not self.redfish_extensions_alias_ok:
                html_str += '<tr><td class="fail log">ERROR - $metadata does not include the required "RedfishExtensions.v1_0_0" namespace with an alias of "Redfish"</td></tr>'
            if len(self.get_missing_namespaces()) > 0:
                html_str += '<tr><td class="fail log">ERROR - The following namespaces are referenced by the service, but are not included in $metadata:<ul>'
                for ns in self.get_missing_namespaces():
                    html_str += '<li>{}</li>'.format(ns)
                html_str += '</ul></td></tr>'
            if len(self.bad_tags) > 0:
                html_str += '<tr><td class="fail log">ERROR - The following tag names in $metadata are unrecognized (check spelling or case):'
                html_str += tag_list_html(self.bad_tags)
                html_str += '</td></tr>'
            if len(self.refs_missing_uri) > 0:
                html_str += '<tr><td class="fail log">ERROR - The following Reference tags in $metadata are missing the expected Uri attribute (check spelling or case):'
                html_str += tag_list_html(self.refs_missing_uri)
                html_str += '</td></tr>'
            if len(self.includes_missing_ns) > 0:
                html_str += '<tr><td class="fail log">ERROR - The following Include tags in $metadata are missing the expected Namespace attribute (check spelling or case):'
                html_str += tag_list_html(self.includes_missing_ns)
                html_str += '</td></tr>'
            if len(self.bad_tag_ns) > 0:
                html_str += '<tr><td class="fail log">ERROR - The following tags in $metadata have an unexpected namespace:'
                html_str += tag_list_html(self.bad_tag_ns)
                html_str += '</td></tr>'
            if len(self.bad_schema_uris) > 0:
                html_str += '<tr><td class="fail log">ERROR - The following schema URIs referenced from $metadata could not be retrieved:'
                html_str += list_html(self.bad_schema_uris)
                html_str += '</td></tr>'
            if len(self.bad_namespace_include) > 0:
                html_str += '<tr><td class="fail log">ERROR - The following namespaces included in $metadata could not be found in the referenced schema URI:'
                html_str += list_html(self.bad_namespace_include)
                html_str += '</td></tr>'
        html_str += '</table>'

        return html_str


class OrderedCounter(Counter, OrderedDict):
    """Counter that remembers the order elements are first encountered"""

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, OrderedDict(self))

    def __reduce__(self):
        return self.__class__, (OrderedDict(self),)

def storeSchemaToLocal(xml_data, origin, service):
    """storeSchemaToLocal

    Moves data pulled from service/online to local schema storage

    Does NOT do so if preferonline is specified

    :param xml_data: data being transferred
    :param origin: origin of xml pulled
    """
    config = service.config
    SchemaLocation = config['metadatafilepath']
    if not os.path.isdir(SchemaLocation):
        os.makedirs(SchemaLocation)
    if 'localFile' not in origin and '$metadata' not in origin:
        __, xml_name = origin.rsplit('/', 1)
        new_file = os.path.join(SchemaLocation, xml_name)
        if not os.path.isfile(new_file):
            with open(new_file, "w") as filehandle:
                filehandle.write(xml_data)
                my_logger.info('Writing online XML to file: {}'.format(xml_name))
        else:
            my_logger.info('NOT writing online XML to file: {}'.format(xml_name))


@lru_cache(maxsize=64)
def getSchemaDetails(service, SchemaType, SchemaURI):
    """
    Find Schema file for given Namespace.

    param SchemaType: Schema Namespace, such as ServiceRoot
    param SchemaURI: uri to grab schema, given LocalOnly is False
    return: (success boolean, a Soup object, origin)
    """
    my_logger.debug('getting Schema of {} {}'.format(SchemaType, SchemaURI))

    if SchemaType is None:
        return False, None, None

    if service is None:
        return getSchemaDetailsLocal(SchemaType, SchemaURI, {})

    elif service.active and getNamespace(SchemaType) in service.metadata.schema_store:
        result = service.metadata.schema_store[getNamespace(SchemaType)]
        if result is not None:
            return True, result.soup, result.origin

    success, soup, origin = getSchemaDetailsLocal(SchemaType, SchemaURI, service.config)
    if success:
        return success, soup, origin

    xml_suffix = '_v1.xml'

    if (SchemaURI is not None) or (SchemaURI is not None and '/redfish/v1/$metadata' in SchemaURI):
        # Get our expected Schema file here
        # if success, generate Soup, then check for frags to parse
        #   start by parsing references, then check for the refLink
        if '#' in SchemaURI:
            base_schema_uri, frag = tuple(SchemaURI.rsplit('#', 1))
        else:
            base_schema_uri, frag = SchemaURI, None
        success, data, response, elapsed = service.callResourceURI(base_schema_uri)
        if success:
            soup = BeautifulSoup(data, "xml")
            # if frag, look inside xml for real target as a reference
            if frag is not None:
                # prefer type over frag, truncated down
                # using frag, check references
                frag = getNamespace(SchemaType)
                frag = frag.split('.', 1)[0]
                refType, refLink = getReferenceDetails(
                    soup, name=base_schema_uri).get(frag, (None, None))
                if refLink is not None:
                    success, linksoup, newlink = getSchemaDetails(service, refType, refLink)
                    if success:
                        return True, linksoup, newlink
                    else:
                        my_logger.error(
                            "SchemaURI couldn't call reference link {} inside {}".format(frag, base_schema_uri))
                else:
                    my_logger.error(
                        "SchemaURI missing reference link {} inside {}".format(frag, base_schema_uri))
                    # error reported; assume likely schema uri to allow continued validation
                    uri = 'http://redfish.dmtf.org/schemas/v1/{}{}'.format(frag, xml_suffix)
                    my_logger.info("Continue assuming schema URI for {} is {}".format(SchemaType, uri))
                    return getSchemaDetails(service, SchemaType, uri)
            else:
                storeSchemaToLocal(data, base_schema_uri, service)
                return True, soup, base_schema_uri
        else:
            my_logger.debug("SchemaURI called unsuccessfully: {}".format(base_schema_uri))
    return getSchemaDetailsLocal(SchemaType, SchemaURI, service.config)


def getSchemaDetailsLocal(SchemaType, SchemaURI, config):
    """
    Find Schema file for given Namespace, from local directory

    param SchemaType: Schema Namespace, such as ServiceRoot
    param SchemaURI: uri to grab schem (generate information from it)
    return: (success boolean, a Soup object, origin)
    """
    Alias = getNamespaceUnversioned(SchemaType)
    SchemaLocation, SchemaSuffix = config['metadatafilepath'], '_v1.xml'
    if SchemaURI is not None:
        uriparse = SchemaURI.split('/')[-1].split('#')
        xml = uriparse[0]
    else:
        my_logger.warning("SchemaURI was empty, must generate xml name from type {}".format(SchemaType)),
        return getSchemaDetailsLocal(SchemaType, Alias + SchemaSuffix, config)
    my_logger.debug(('local', SchemaType, SchemaURI, SchemaLocation + '/' + xml))
    filestring = Alias + SchemaSuffix if xml is None else xml
    try:
        # get file
        with open(SchemaLocation + '/' + xml, "r") as filehandle:
            data = filehandle.read()

        # get tags
        soup = BeautifulSoup(data, "xml")
        edmxTag = soup.find('Edmx', recursive=False)
        parentTag = edmxTag.find('DataServices', recursive=False)
        child = parentTag.find('Schema', recursive=False)
        SchemaNamespace = child['Namespace']
        FoundAlias = SchemaNamespace.split(".")[0]
        my_logger.debug(FoundAlias)

        if FoundAlias in Alias:
            return True, soup, "localFile:" + SchemaLocation + '/' + filestring

    except FileNotFoundError:
        # if we're looking for $metadata locally... ditch looking for it, go straight to file
        if '/redfish/v1/$metadata' in SchemaURI and Alias != '$metadata':
            my_logger.debug("Unable to find a xml of {} at {}, defaulting to {}".format(SchemaURI, SchemaLocation, Alias + SchemaSuffix))
            return getSchemaDetailsLocal(SchemaType, Alias + SchemaSuffix, config)
        else:
            my_logger.warning("Schema file {} not found in {}".format(filestring, SchemaLocation))
            if Alias == '$metadata':
                my_logger.warning("If $metadata cannot be found, Annotations may be unverifiable")
    except Exception as ex:
        my_logger.error("A problem when getting a local schema has occurred {}".format(SchemaURI))
        my_logger.warning("output: ", exc_info=True)
    return False, None, None


def check_redfish_extensions_alias(name, namespace, alias):
    """
    Check that edmx:Include for Namespace RedfishExtensions has the expected 'Redfish' Alias attribute
    :param name: the name of the resource
    :param item: the edmx:Include item for RedfishExtensions
    :return: bool
    """
    if alias is None or alias != 'Redfish':
        msg = ("In the resource {}, the {} namespace must have an alias of 'Redfish'. The alias is {}. " +
               "This may cause properties of the form [PropertyName]@Redfish.TermName to be unrecognized.")
        my_logger.error(msg.format(name, namespace,
                             'missing' if alias is None else "'" + str(alias) + "'"))
        return False
    return True


def getReferenceDetails(soup, metadata_dict=None, name='xml'):
    """
    Create a reference dictionary from a soup file

    param arg1: soup
    param metadata_dict: dictionary of service metadata, compare with
    return: dictionary
    """
    includeTuple = namedtuple('include', ['Namespace', 'Uri'])
    refDict = {}

    maintag = soup.find("Edmx", recursive=False)
    reftags = maintag.find_all('Reference', recursive=False)
    for ref in reftags:
        includes = ref.find_all('Include', recursive=False)
        for item in includes:
            uri = ref.get('Uri')
            ns, alias = (item.get(x) for x in ['Namespace', 'Alias'])
            if ns is None or uri is None:
                my_logger.error("Reference incorrect for: {}".format(item))
                continue
            if alias is None:
                alias = ns
            refDict[alias] = includeTuple(ns, uri)
            # Check for proper Alias for RedfishExtensions
            if name == '$metadata' and ns.startswith('RedfishExtensions.'):
                check_bool = check_redfish_extensions_alias(name, ns, alias)

    cntref = len(refDict)
    if metadata_dict is not None:
        refDict.update(metadata_dict)
    my_logger.debug("METADATA: References generated from {}: {} out of {}".format(name, cntref, len(refDict)))
    return refDict


# Copyright Notice:
# Copyright 2016-2019 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

"""
Redfish Service Validator GUI

File : RedfishServiceValidatorGui.py

Brief : This file contains the GUI to interact with the RedfishServiceValidator
"""

import configparser
import os
import threading
import tkinter as tk
from tkinter import filedialog as tkFileDialog
import traceback
import webbrowser

import redfish_service_validator.RedfishLogo as logo
import redfish_service_validator.RedfishServiceValidator as rsv

g_config_file_name = "config/config.ini"

g_config_defaults = {
    "Tool": {
        "verbose": {
            "value": "",
            "description": "Level of verbosity (0-3)"
        }
    },
    "Host": {
        "ip": {
            "value": "http://localhost:8000",
            "description": "Host of testing system, formatted as https:// ip : port (can use http as well)"
        },
        "username": {
            "value": "MyUser",
            "description": "Username for Basic authentication"
        },
        "password": {
            "value": "MyPass",
            "description": "Username for Basic authentication"
        },
        "description": {
            "value": "MySystem",
            "description": "Description of system being tested (optional)"
        },
        "forceauth": {
            "value": "False",
            "description": "Force authentication even on http servers"
        },
        "authtype": {
            "value": "Basic",
            "description": "Authorization type (Basic | Session | Token | None)"
        },
        "token": {
            "value": "False",
            "description": "Token string for Token authentication"
        },
        "ext_http_proxy": {
            "value": "",
            "description": "URL of the HTTP proxy for accessing external sites"
        },
        "ext_https_proxy": {
            "value": "",
            "description": "URL of the HTTPS proxy for accessing external sites"
        },
        "serv_http_proxy": {
            "value": "",
            "description": "URL of the HTTP proxy for accessing the service"
        },
        "serv_https_proxy": {
            "value": "",
            "description": "URL of the HTTPS proxy for accessing the service"
        }
    },
    "Validator": {
        "payload": {
            "value": "",
            "description": "Option to test a specific payload or resource tree (see README)"
        },
        "logdir": {
            "value": "./logs",
            "description": "Place to save logs and run configs"
        },
        "oemcheck": {
            "value": "True",
            "description": "Whether to check Oem items on service"
        },
        "debugging": {
            "value": "False",
            "description": "Whether to print debug to log"
        },
        "uricheck": {
            "value": "False",
            "description": "Whether to force urichecking if under RedfishVersion 1.6.0"
        },
        "schema_directory": {
            "value": "./SchemaFiles/metadata",
            "description": "Where schema is located/saved on system"
        },
        "mockup": {
            "value": "",
            "description": "Enables insertion of local mockup resources to replace payloads from the service"
        }
    }
}

class RSVGui:
    """
    Main class for the GUI

    Args:
        parent (Tk): Parent Tkinter object
    """

    def __init__( self, parent ):
        # Set up the configuration
        self.config = {}
        for section in g_config_defaults:
            self.config[section] = {}
            for option in g_config_defaults[section]:
                self.config[section][option] = g_config_defaults[section][option]

        # Read in the config file, and apply any valid settings
        self.config_file = g_config_file_name
        self.system_under_test = tk.StringVar()
        self.parse_config()

        # Initialize the window
        self.parent = parent
        self.parent.title( "Redfish Service Validator {}".format( rsv.tool_version ) )

        # Add the menu bar
        menu_bar = tk.Menu( self.parent )
        file_menu = tk.Menu( menu_bar, tearoff = 0 )
        file_menu.add_command( label = "Open Config", command = self.open_config )
        file_menu.add_command( label = "Save Config", command = self.save_config )
        file_menu.add_command( label = "Save Config As", command = self.save_config_as )
        file_menu.add_command( label = "Edit Config", command = self.edit_config )
        file_menu.add_separator()
        file_menu.add_command( label = "Exit", command = self.parent.destroy )
        menu_bar.add_cascade( label = "File", menu = file_menu )
        self.parent.config( menu = menu_bar )

        # Add the logo
        image = tk.PhotoImage( data = logo.logo )
        label = tk.Label( self.parent, image = image, width = 384 )
        label.image = image
        label.pack( side = tk.TOP )

        # Add the system under test label
        tk.Label( self.parent, textvariable = self.system_under_test, font = ( None, 12 ) ).pack( side = tk.TOP )

        # Add the buttons
        button_frame = tk.Frame( self.parent )
        button_frame.pack( side = tk.TOP, fill = tk.X )
        self.run_button_text = tk.StringVar()
        self.run_button_text.set( "Run Test" )
        self.run_button = tk.Button( button_frame, textvariable = self.run_button_text, command = self.run )
        self.run_button.pack( side = tk.LEFT )
        self.run_label_text = tk.StringVar()
        self.run_label_text.set( "" )
        tk.Label( button_frame, textvariable = self.run_label_text ).pack( side = tk.LEFT )
        tk.Button( button_frame, text = "Exit", command = self.parent.destroy ).pack( side = tk.RIGHT )

    def update_sut( self ):
        """
        Updates the System Under Test string
        """
        self.system_under_test.set( "System Under Test: " + self.config["Host"]["ip"]["value"] )

    def parse_config( self ):
        """
        Parses the configuration settings from a file
        """
        config_parser = configparser.ConfigParser()
        config_parser.optionxform = str
        config_parser.read( self.config_file )
        for section in config_parser.sections():
            for option in config_parser.options( section ):
                if section in self.config:
                    if option in self.config[section]:
                        self.config[section][option]["value"] = config_parser.get( section, option )
        self.update_sut()

    def build_config_parser( self, preserve_case ):
        """
        Builds a config parser element from the existing configuration

        Args:
            preserve_case (bool): True if the casing of the options is to be preserved

        Returns:
            ConfigParser: A ConfigParser object generated from the configuration data
        """
        config_parser = configparser.ConfigParser()
        if preserve_case:
            config_parser.optionxform = str
        for section in self.config:
            config_parser.add_section( section )
            for option in self.config[section]:
                config_parser.set( section, option, self.config[section][option]["value"] )
        return config_parser

    def open_config( self ):
        """
        Opens the configuration settings from a file
        """
        filename = tkFileDialog.askopenfilename( initialdir = os.getcwd(), title = "Open", filetypes = ( ( "INI", "*.ini" ), ( "All Files", "*.*" ) ) )
        if filename == "":
            # User closed the box; just return
            return
        self.config_file = filename
        self.parse_config()

    def edit_config( self ):
        """
        Edits the configuration settings
        """
        option_win = tk.Toplevel()
        option_win_frame = tk.Frame( option_win )
        option_win_canvas = tk.Canvas( option_win_frame )
        option_y_scroll = tk.Scrollbar( option_win_frame, orient = "vertical", command = option_win_canvas.yview )
        option_y_scroll.pack( side = tk.RIGHT, fill = tk.Y )
        option_x_scroll = tk.Scrollbar( option_win, orient = "horizontal", command = option_win_canvas.xview )
        option_x_scroll.pack( side = tk.BOTTOM, fill = tk.X )
        option_win_frame.pack( side = tk.TOP, fill = tk.BOTH, expand = True )
        option_win_canvas.pack( side = tk.LEFT, fill = tk.BOTH, expand = True )
        option_win_canvas.bind( "<Configure>", lambda e: option_win_canvas.configure( scrollregion = option_win_canvas.bbox( "all" ) ) )
        option_win_contents = tk.Frame( option_win_canvas )
        option_win_canvas.create_window( ( 0, 0 ), window = option_win_contents )
        config_values = {}

        # Iterate through the config file options to build the window
        for section in self.config:
            config_values[section] = {}
            section_frame = tk.Frame( option_win_contents )
            section_frame.pack( side = tk.TOP )
            tk.Label( section_frame, text = section, anchor = "center", font = ( None, 16 ) ).pack( side = tk.LEFT )
            for option in self.config[section]:
                option_frame = tk.Frame( option_win_contents )
                option_frame.pack( side = tk.TOP, fill = tk.X )
                tk.Label( option_frame, text = option, width = 16, anchor = "w" ).pack( side = tk.LEFT )
                config_values[section][option] = tk.StringVar()
                config_values[section][option].set( self.config[section][option]["value"] )
                if "options" in self.config[section][option]:
                    option_menu = tk.OptionMenu( option_frame, config_values[section][option], *self.config[section][option]["options"] )
                    option_menu.configure( width = 26 )    # Need a better way to fine tune this so it lines up nicely with the text boxes
                    option_menu.pack( side = tk.LEFT )
                else:
                    tk.Entry( option_frame, width = 32, textvariable = config_values[section][option] ).pack( side = tk.LEFT )
                tk.Label( option_frame, text = self.config[section][option]["description"], anchor = "w" ).pack( side = tk.LEFT )
        tk.Button( option_win_contents, text = "Apply", command = lambda: self.apply_config( option_win, config_values ) ).pack( side = tk.BOTTOM )
        option_win_contents.update()
        option_win_canvas.config( xscrollcommand = option_x_scroll.set, yscrollcommand = option_y_scroll.set, width = option_win_contents.winfo_width(), height = option_win_contents.winfo_height() )

    def apply_config( self, window, config_values ):
        """
        Applies the configation settings from the edit window

        Args:
            window (Toplevel): Tkinter Toplevel object with text boxes to apply
            config_values (Array): An array of StringVar objects with the user input
        """
        for section in self.config:
            for option in self.config[section]:
                self.config[section][option]["value"] = config_values[section][option].get()
        self.update_sut()
        window.destroy()

    def save_config( self ):
        """
        Saves the config file
        """
        config_parser = self.build_config_parser( True )
        with open( self.config_file, "w" ) as config_file:
            config_parser.write( config_file )

    def save_config_as( self ):
        """
        Saves the config file as a new file
        """
        filename = tkFileDialog.asksaveasfilename( initialdir = os.getcwd(), title = "Save As", filetypes = ( ( "INI", "*.ini" ), ( "All Files", "*.*" ) ) )
        if filename == "":
            # User closed the box; just return
            return
        self.config_file = filename
        if not self.config_file.lower().endswith( ".ini" ):
            self.config_file = self.config_file + ".ini"
        self.save_config()

    def run( self ):
        """
        Runs the service validator
        """
        self.run_button_text.set( "Running" )
        self.run_button.config( state = tk.DISABLED )
        run_thread = threading.Thread( target = self.run_imp )
        run_thread.daemon = True
        run_thread.start()

    def run_imp( self ):
        """
        Thread for running the service validator so the GUI doesn't freeze
        """
        self.run_label_text.set( "Test running; please wait" )

        run_window = tk.Toplevel()
        run_text_frame = tk.Frame( run_window )
        run_text_frame.pack( side = tk.TOP )
        run_scroll = tk.Scrollbar( run_text_frame )
        run_scroll.pack( side = tk.RIGHT, fill = tk.Y )
        run_text = tk.Text( run_text_frame, height = 48, width = 128, yscrollcommand = run_scroll.set )
        rsv.my_logger.handlers[0].stream = RunOutput( run_text )
        run_text.pack( side = tk.TOP )
        run_button_frame = tk.Frame( run_window )
        run_button_frame.pack( side = tk.BOTTOM )
        tk.Button( run_button_frame, text = "OK", command = run_window.destroy ).pack( side = tk.LEFT )
        tk.Button( run_button_frame, text = "Copy", command = lambda: self.copy_text( run_text ) ).pack( side = tk.RIGHT )

        # Launch the validator
        try:
            rsv_config = self.build_config_parser( False )
            status_code, last_results_page, exit_string = rsv.validate( configfile = rsv_config )
            if last_results_page is not None:
                webbrowser.open_new( last_results_page )
            else:
                # The validation could not take place (for a controlled reason)
                notification_window = tk.Toplevel()
                tk.Label( notification_window, text = "Test aborted: " + exit_string, anchor = "center" ).pack( side = tk.TOP )
                tk.Button( notification_window, text = "OK", command = notification_window.destroy ).pack( side = tk.BOTTOM )
        except:
            oops_window = tk.Toplevel()
            tk.Label( oops_window, text = "Please copy the info below and file an issue on GitHub!", width = 64, anchor = "center" ).pack( side = tk.TOP )
            oops_text_frame = tk.Frame( oops_window )
            oops_text_frame.pack( side = tk.TOP )
            oops_scroll = tk.Scrollbar( oops_text_frame )
            oops_scroll.pack( side = tk.RIGHT, fill = tk.Y )
            oops_text = tk.Text( oops_text_frame, height = 32, width = 64, yscrollcommand = oops_scroll.set )
            oops_text.insert( tk.END, traceback.format_exc() )
            oops_text.pack( side = tk.TOP )
            oops_button_frame = tk.Frame( oops_window )
            oops_button_frame.pack( side = tk.BOTTOM )
            tk.Button( oops_button_frame, text = "OK", command = oops_window.destroy ).pack( side = tk.LEFT )
            tk.Button( oops_button_frame, text = "Copy", command = lambda: self.copy_text( oops_text ) ).pack( side = tk.RIGHT )
        self.run_button.config( state = tk.NORMAL )
        self.run_button_text.set( "Run Test" )
        self.run_label_text.set( "Test Complete" )

    def copy_text( self, text ):
        """
        Copies text to the system clipboard

        Args:
            text (Text): Tkinter Text object with text to copy
        """
        self.parent.clipboard_clear()
        self.parent.clipboard_append( text.get( 1.0, tk.END ) )

class RunOutput( object ):
    """
    Runtime output class

    Args:
        text (Text): Tkinter Text object to use as the output
    """

    def __init__( self, text ):
        self.output = text

    def write( self, string ):
        """
        Writes to the output object

        Args:
            string (string): The string to output
        """
        if self.output.winfo_exists():
            self.output.insert( tk.END, string )
            self.output.see( tk.END )

def main():
    """
    Entry point for the GUI
    """
    root = tk.Tk()
    RSVGui( root )
    root.mainloop()

if __name__ == '__main__':
    main()



import glob, copy, difflib
import logging
import re
from collections import namedtuple
from enum import Enum, auto
from os import path

from bs4 import BeautifulSoup

from redfish_service_validator.helper import (
    getNamespace,
    getNamespaceUnversioned,
    getType,
    getVersion,
    splitVersionString,
    stripCollection
)

includeTuple = namedtuple("include", ["Namespace", "Uri"])

my_logger = logging.getLogger(__name__)

REDFISH_ABSENT = "n/a"

URI_ID_REGEX = '\{[A-Za-z0-9]*Id\}'

VALID_ID_REGEX = '([A-Za-z0-9.!#$&-;=?\[\]_~])+'


# Excerpt definitions
class ExcerptTypes(Enum):
    NEUTRAL = auto()
    CONTAINS = auto()
    ALLOWED = auto()
    EXCLUSIVE = auto()


excerpt_info_by_type = {
    'Redfish.ExcerptCopy': ExcerptTypes.CONTAINS,
    'Redfish.Excerpt': ExcerptTypes.ALLOWED,
    'Redfish.ExcerptCopyOnly': ExcerptTypes.EXCLUSIVE
}


allowed_annotations = ['odata', 'Redfish', 'Privileges', 'Message']


def get_fuzzy_property(prop_name: str, jsondata: dict, allPropList=[]):
    """
    Get property closest to the discovered property.

    Args:
        prop_name (str): Key of property
        jsondata (dict): Dictionary of payload
        allPropList (list, optional): List of possible properties of this particular payload. Defaults to [].

    Returns:
        prop_name: Closest match
        rtype: str
    """
    possibleMatch = difflib.get_close_matches(prop_name, list(jsondata), 1, 0.70)
    if len(possibleMatch) > 0 and possibleMatch[0] not in [
        s[2] for s in allPropList if s[2] != prop_name
    ]:
        val = jsondata.get(possibleMatch[0], "n/a")
        if val != "n/a":
            return possibleMatch[0]
            # .error('{} was not found in payload, attempting closest match: {}'.format(newProp, pname))
    return prop_name


class MissingSchemaError(Exception):
    """
    Missing Schema Error.
    
    Raise when Catalog is unable to find a Schema
    """

    pass


class SchemaCatalog:
    """
    Catalog for holding Schema files.
    
    From Catalog, you can get any Schema by it's filename, or its classes
    """
    # TODO: Generate documents on demand, not all at once

    def __init__(self, filepath: str, metadata: object = None):
        """Init

        Args:
            filepath (str): Directory of metadata
            metadata (object, optional): Preestablished metadata. Defaults to None.
        """
        self.filepath = filepath
        self.alias = {}
        self.catalog = {}
        self.catalog_by_class = {}
        self.flags = {
            'ignore_uri_checks': False
        }
        my_logger.debug("Creating Schema catalog from filepath {}".format(filepath))

        # create SchemaDoc objects
        for x in glob.glob(path.join(filepath, "*")):
            with open(x) as f:
                my_name = path.split(x)[-1]
                schema = SchemaDoc(f.read(), self, my_name)
                self.catalog[my_name] = schema

            base_names = [getNamespaceUnversioned(x) for x in schema.classes if getNamespaceUnversioned(x) not in schema.classes]
            for item in list(schema.classes.keys()) + base_names:
                if item not in self.catalog_by_class:
                    self.catalog_by_class[item] = [schema]
                else:
                    self.catalog_by_class[item].append(schema)
        
        for _, schema in self.catalog.items():
            self.alias.update(schema.alias)

    def getSchemaDocByClass(self, typename):
        """
        Get Document by class

        :param typename: type string
        :type typename: str
        :raises MissingSchemaError: Missing schema in Catalog
        :return: Schema Document
        :rtype: SchemaDoc
        """
        typename = getNamespaceUnversioned(stripCollection(typename))
        typename = self.alias.get(typename, typename)
        if typename in self.catalog_by_class:
            return self.catalog_by_class[typename][0]
        else:
            raise MissingSchemaError( "Could not find any Schema with these parameters {}".format(typename))

    def getSchemaInCatalog(self, typename):
        """
        Get Schema by class

        :param typename: type string
        :type typename: str
        :raises MissingSchemaError: Missing schema in Catalog
        :return: Schema Document
        :rtype: SchemaClass
        """

        typename = getNamespace(typename)
        typename = self.alias.get(typename, typename)
        my_doc = self.getSchemaDocByClass(typename)
        return my_doc.classes[typename]

    def getTypeInCatalog(self, fulltype):
        """
        Get Type via type string

        :param typename: type string
        :type typename: str
        :raises MissingSchemaError: Missing schema in Catalog
        :return: Schema Document
        :rtype: RedfishType
        """
        typename = getNamespaceUnversioned(fulltype)
        typename = self.alias.get(typename, typename)
        my_doc = self.getSchemaDocByClass(typename)
        return my_doc.getTypeInSchemaDoc(fulltype)
        

class SchemaDoc:
    """Represents a schema document."""

    def __init__(self, data: str, catalog: SchemaCatalog = None, name: str = None):
        # set up document
        self.soup = BeautifulSoup(data, "xml")
        self.name = str(name)
        self.origin = "local"
        self.catalog = catalog
        self.classes = {}
        self.alias = {}

        edmxTag = self.soup.find("Edmx", recursive=False)
        reftags = edmxTag.find_all("Reference", recursive=False)
        self.refs = {}
        for ref in reftags:
            includes = ref.find_all("Include", recursive=False)
            for item in includes:
                uri = ref.get("Uri")
                ns, alias = (item.get(x) for x in ["Namespace", "Alias"])
                if ns is None or uri is None:
                    my_logger.error("Reference incorrect for: {}".format(item))
                    continue
                if alias is None:
                    alias = ns
                else:
                    self.alias[alias] = ns

                self.refs[alias] = includeTuple(ns, uri)
                # Check for proper Alias for RedfishExtensions
                # if name == '$metadata' and ns.startswith('RedfishExtensions.'):
                #     check_bool = check_redfish_extensions_alias(name, ns, alias)

        cntref = len(self.refs)

        parentTag = edmxTag.find("DataServices", recursive=False)
        children = parentTag.find_all("Schema", recursive=False)
        self.classes = {}
        for child in children:
            self.classes[child["Namespace"]] = SchemaClass(child, self)
        my_logger.debug(
            "References generated from {}: {} out of {}".format(
                name, cntref, len(self.refs)
            )
        )

    def getReference(self, namespace):
        """getSchemaFromReference

        Get tuple from generated references

        :param namespace: Namespace of reference
        """
        tup = self.refs.get(namespace)
        tupVersionless = self.refs.get(getNamespace(namespace))
        if tup is None:
            if tupVersionless is None:
                my_logger.warning('No such reference {} in {}, returning None...'.format(namespace, self.origin))
                return None
            else:
                tup = tupVersionless
                my_logger.warning('No such reference {} in {}, using unversioned...'.format(namespace, self.origin))
        return tup

    def getTypeInSchemaDoc(self, currentType, tagType=["EntityType", "ComplexType"]):
        """getTypeTagInSchema

        Get type tag in schema

        :param currentType: type string
        :param tagType: Array or single string containing the xml tag name
        
        rtype: SchemaClass
        """
        if isinstance(currentType, RedfishType):
            currentType = currentType.fulltype

        currentType = stripCollection(currentType)

        pnamespace, ptype = getNamespace(currentType), getType(currentType)
        pnamespace = self.catalog.alias.get(pnamespace, pnamespace)
        pbase = getNamespaceUnversioned(pnamespace)

        if pnamespace not in self.classes:
            if pbase in self.classes:
                my_logger.error("Namespace of type {} appears missing from SchemaXML {}, attempting highest type: {}".format(pnamespace, self.name, currentType))
                ns_obj = self.classes[pbase]
                pnamespace = getNamespace(ns_obj.getHighestType(currentType))
                my_logger.error("New namespace: {}".format(pnamespace))
            elif pnamespace in self.refs or pbase in self.refs:
                new_doc = self.catalog.getSchemaDocByClass(pnamespace)
                return new_doc.getTypeInSchemaDoc(currentType)
        if pnamespace in self.classes:
            currentNamespace = self.classes[pnamespace]
            return currentNamespace.my_types[ptype]
        else:
            my_logger.error("Namespace of type {} appears missing from SchemaXML {}...".format(pnamespace, self.name))
            raise MissingSchemaError('No such schema referenced in this document')
    

class SchemaClass:
    def __init__(self, soup, owner: SchemaDoc):
        super().__init__()
        self.parent_doc = owner
        self.catalog = owner.catalog
        self.class_soup = soup
        self.class_name = soup["Namespace"]
        self.entity_types, self.complex_types, self.enum_types, self.def_types = {}, {}, {}, {}

        for x in self.class_soup.find_all(["EntityType"], recursive=False):
            self.entity_types[x["Name"]] = RedfishType(x, self)

        for x in self.class_soup.find_all(["ComplexType"], recursive=False):
            self.complex_types[x["Name"]] = RedfishType(x, self)

        for x in self.class_soup.find_all(["EnumType"], recursive=False):
            self.enum_types[x["Name"]] = RedfishType(x, self)

        for x in self.class_soup.find_all(["TypeDefinition"], recursive=False):
            self.def_types[x["Name"]] = RedfishType(x, self)
        
        self.actions = {}
        for x in self.class_soup.find_all(["Action"], recursive=False):
            self.actions[x["Name"]] = x

        self.terms = {}
        for x in self.class_soup.find_all(["Term"], recursive=False):
            self.terms[x["Name"]] = RedfishType(x, self)

        self.my_types = {**self.entity_types, **self.complex_types, **self.enum_types, **self.def_types}

    def getHighestType(self, my_full_type, limit=None):
        """
        Get Highest possible version for given type.

        :param acquiredtype: Type available
        :param limit: Version string limit (full namespace or just version 'v1_x_x')
        """
        typelist = []
        my_type = getType(my_full_type)

        if limit is not None:
            if getVersion(limit) is None:
                # logger.warning('Limiting namespace has no version, erasing: {}'.format(limit))
                limit = None
            else:
                limit = getVersion(limit)

        for _, schema in self.parent_doc.classes.items():
            newNamespace = schema.class_name
            if limit is not None:
                if getVersion(newNamespace) is None:
                    continue
                if splitVersionString(newNamespace) > splitVersionString(limit):
                    continue
            if (my_type in self.my_types):
                typelist.append(splitVersionString(newNamespace))

        if len(typelist) > 1:
            for ns in reversed(sorted(typelist)):
                my_logger.debug("{}   {}".format(ns, getType(my_type)))
                my_type = getNamespaceUnversioned(my_full_type) + ".v{}_{}_{}".format(*ns) + "." + getType(my_full_type)
                return my_type
        return my_type


class RedfishType:
    """Redfish Type

    Represents tags of 'Property', 'NavigationProperty' in an EntityType/ComplexType
    And also represents EntityType/ComplexType/EnumType/TypeDefinitions, not basic types like Edm
    """
    def __eq__(self, other):
        if isinstance(other, str):
            return other == str(self)
        else:
            return self == other

    def __repr__(self):
        return self.fulltype
    
    def __init__(self, soup, owner: SchemaClass):

        self.owner = owner
        self.catalog = owner.catalog

        self.type_soup = soup
        self.tag_type = soup.name

        if self.tag_type in ['NavigationProperty', 'Property', 'Term']:
            self.IsPropertyType = True
            self.IsNav = self.tag_type in ['NavigationProperty']
            self.fulltype = self.parent_type
        else:
            self.IsPropertyType = False
            self.IsNav = False
            self.fulltype = self.owner.class_name + '.' + soup['Name']

        if 'Collection(' in self.fulltype:
            my_fulltype = self.fulltype.replace('Collection(', "").replace(')', "")
            self.Namespace, self.TypeName = getNamespace(my_fulltype), getType(my_fulltype)
        else:
            self.Namespace, self.TypeName = getNamespace(self.fulltype), getType(self.fulltype)

        self.tags = {}
        for tag in self.type_soup.find_all(recursive=False):
            if tag.get('Term'):
                self.tags[tag['Term']] = tag.attrs
                if (tag.get('Term') == 'Redfish.Revisions'):
                    self.tags[tag['Term']] = tag.find_all('Record')

        propPermissions = self.tags.get('OData.Permissions')

        self.IsMandatory = self.tags.get('Redfish.Required') is not None
        self.IsNullable = self.type_soup.get("Nullable", "true") not in ["false", False, "False"]
        self.AutoExpand = (self.tags.get('OData.AutoExpand') or self.tags.get('OData.AutoExpand'.lower())) is not None
        self.Deprecated = self.tags.get('Redfish.Deprecated')
        self.Revisions = self.tags.get('Redfish.Revisions')
        self.Excerpt = False

        self.Permissions = propPermissions['EnumMember'] if propPermissions is not None else None

        self.excerptType = ExcerptTypes.NEUTRAL
        self.excerptTags = []

        for annotation, val in excerpt_info_by_type.items():
            if annotation in self.tags:
                self.excerptTags = self.tags.get(annotation).get('String', '').split(',')
                self.excerptTags = [x.strip(' ') for x in self.excerptTags] if self.excerptTags != [''] else []
                self.excerptType = val
        
        self.Excerpt = self.excerptType != ExcerptTypes.NEUTRAL

        self.property_pattern = None

        # get properties
        prop_tags = self.type_soup.find_all( ["NavigationProperty", "Property"], recursive=False)
    
        self.unique_properties = {}

        for innerelement in prop_tags:
            prop_name = innerelement["Name"]
            self.unique_properties[prop_name] = RedfishType(innerelement, self.owner)
    
    @property
    def HasAdditional(self):
        my_parents = self.getTypeTree()
        for my_type in my_parents:
            if not isinstance(my_type, RedfishType): continue
            if 'Bios' in str(my_type) and 'Attributes' in str(my_type):
                return True
            if my_type == 'MessageRegistry.v1_0_0.MessageProperty':
                return True
            additionalElement = my_type.type_soup.find("Annotation", attrs={"Term": "OData.AdditionalProperties"})
            HasAdditional = ( False if additionalElement is None else (
                    True if additionalElement.get("Bool", False) in ["True", "true", True]
                    else False))
            if HasAdditional:  
                return True
        return False

    @property
    def CanUpdate(self):
        return self.getCapabilities()['CanUpdate']

    @property
    def CanDelete(self):
        return self.getCapabilities()['CanDelete']

    @property
    def CanInsert(self):
        return self.getCapabilities()['CanInsert']

    def getCapabilities(self):
        my_dict = {'CanUpdate': False,
                   'CanInsert': False,
                   'CanDelete': False}

        my_parents = self.getTypeTree()
        for my_type in reversed(my_parents):
            if not isinstance(my_type, RedfishType): continue
            try:
                element = my_type.type_soup.find("Annotation", attrs={"Term": "Capabilities.InsertRestrictions"})
                if element:
                    my_dict['CanInsert'] = element.find("PropertyValue").get('Bool', 'False').lower() == 'true'
                element = my_type.type_soup.find("Annotation", attrs={"Term": "Capabilities.UpdateRestrictions"})
                if element:
                    my_dict['CanUpdate'] = element.find("PropertyValue").get('Bool', 'False').lower() == 'true'
                element = my_type.type_soup.find("Annotation", attrs={"Term": "Capabilities.DeleteRestrictions"})
                if element:
                    my_dict['CanDelete'] = element.find("PropertyValue").get('Bool', 'False').lower() == 'true'
            except Exception as e:
                my_logger.debug('Exception caught while checking Uri', exc_info=1)
                my_logger.warning('Could not gather info from Capabilities annotation')
                return {'CanUpdate': False, 'CanInsert': False, 'CanDelete': False}

        return my_dict

    @property
    def DynamicProperties(self):
        my_parents = self.getTypeTree()
        for my_type in reversed(my_parents):
            if not isinstance(my_type, RedfishType): continue
            try:
                dynamic = my_type.type_soup.find("Annotation", attrs={"Term": "Redfish.DynamicPropertyPatterns"})
                if dynamic: 
                    # create PropertyPattern dict containing pattern and type for DynamicPropertyPatterns validation
                    pattern_elem = dynamic.find("PropertyValue", Property="Pattern")
                    type_elem = dynamic.find("PropertyValue", Property="Type")
                    if pattern_elem and type_elem:
                        return {
                            "Pattern": pattern_elem.get("String"),
                            "Type": type_elem.get("String"),
                        }
                    if pattern_elem and not type_elem or type_elem and not pattern_elem:
                        raise ValueError('Cannot have pattern with Type in DynamicProperty annotation')

            except Exception as e:
                my_logger.debug('Exception caught while checking Dynamic', exc_info=1)
                my_logger.warning('Could not gather info from DynamicProperties annotation')
                return None 
        
        return None

    def getUris(self):
        """
        Return Redfish.Uris annotation values

        :return: Array of Uris
        :rtype: list
        """
        my_parents = self.getTypeTree()
        expectedUris = []
        for my_type in my_parents:
            if not isinstance(my_type, RedfishType): continue
            uriElement = my_type.type_soup.find("Annotation", attrs={"Term": "Redfish.Uris"})
            if uriElement is not None:
                try:
                    all_strings = uriElement.find("Collection").find_all("String")
                    expectedUris += [e.contents[0] for e in all_strings]
                except Exception as e:
                    my_logger.debug('Exception caught while checking Uri', exc_info=1)
                    my_logger.warning('Could not gather info from Redfish.Uris annotation')
                    expectedUris = []
        return expectedUris

    @property
    def parent_type(self):
        """
        Returns string of the parent type

        Returns:
            string of type
        """
        soup = self.type_soup
        parent_type = (
            soup["UnderlyingType"] if self.tag_type == "TypeDefinition"
            else soup.get("BaseType", soup.get("Type", None))
        )
        return parent_type
        
    def getTypeTree(self):
        """
        Returns tree of RedfishType/string of parent types
        """
        tree = [self]
        my_type = self.parent_type
        while my_type:
            if 'Edm.' not in my_type:
                type_obj = self.owner.parent_doc.catalog.getSchemaDocByClass(my_type).getTypeInSchemaDoc(my_type)
                tree.append(type_obj)
                my_type = type_obj.parent_type
            else:
                return tree + [my_type]
        return tree

    def getBaseType(self):
        """
        Returns string representing our tag type, and if that type is a collection

        TODO:  This should be an enum
            possible values ["complex", "enum", "entity", "Edm.String", "Edm.Int"...]

        Returns:
            string, boolean
            None, False
        """
        if self.tag_type == "EnumType":
            return 'enum'
        if self.tag_type == "ComplexType":
            return 'complex'
        if self.tag_type == "EntityType":
            return 'entity'

        my_type = self.parent_type
        if 'Edm.' in my_type:
            return my_type
        if self.IsPropertyType:
            type_obj = self.owner.parent_doc.catalog.getSchemaDocByClass(my_type).getTypeInSchemaDoc(my_type)
            return type_obj.getBaseType()
        return 'none'
    
    def IsCollection(self):
        tree = [self.fulltype]
        my_type = self.parent_type
        while my_type:
            if isinstance(my_type, RedfishType):
                tree.append(my_type.fulltype)
                my_type = self.parent_type
            else:
                tree.append(my_type)
                break
        return any([re.match(r'Collection\((.*)', typ) for typ in tree])
    
    def getCollectionType(self):
        my_new_type = stripCollection(self.fulltype)
        new_type_obj = self.catalog.getSchemaDocByClass(getNamespace(my_new_type)).getTypeInSchemaDoc(my_new_type)
        return new_type_obj

    def getProperties(self):
        """
        Returns all our properties from our current type and its parents
        """
        all_properties = {}
        for type_obj in self.getTypeTree():
            all_properties.update(type_obj.unique_properties)
        return all_properties

    def validate(self, val, added_pattern=None):
        """
        Returns True if validation succeeds, else raises a ValueError
        """
        my_logger.debug((self, val, self.fulltype, self.tag_type, self.parent_type))
        if val == REDFISH_ABSENT:
            if self.IsMandatory:
                raise ValueError("Should not be absent")
            else:
                return True
        if val is None:
            if not self.IsNullable:
                raise ValueError("Should not be null")
            else:
                return True
        # recurse parent_types until we get a basic type...
        if self.tag_type == "EnumType":
            my_enums = [x["Name"] for x in self.type_soup.find_all("Member")]
            if val not in my_enums:
                raise ValueError("Value {} Enum not found in {}".format(val, my_enums))
        if self.tag_type == "ComplexType":
            if not isinstance(val, dict):
                raise ValueError("Complex value is not Dict")
            my_complex = self.createObject().populate(val)
        if self.tag_type == "EntityType":
            return True
        my_type = self.parent_type
        if my_type:
            if 'Edm.' not in my_type:
                type_obj = self.owner.parent_doc.catalog.getSchemaDocByClass(my_type).getTypeInSchemaDoc(my_type)
                return type_obj.validate(val)
            else:
                enum_annotation = self.type_soup.find('Annotation', attrs={'Term': 'Redfish.Enumeration'}, recursive=False)
                validPatternAttr = self.type_soup.find('Annotation', attrs={'Term': 'Validation.Pattern'})
                validMinAttr = self.type_soup.find('Annotation', attrs={'Term': 'Validation.Minimum'})
                validMaxAttr = self.type_soup.find('Annotation', attrs={'Term': 'Validation.Maximum'})
                validMin, validMax = int(validMinAttr['Int']) if validMinAttr is not None else None, \
                    int(validMaxAttr['Int']) if validMaxAttr is not None else None
                validPattern = validPatternAttr.get('String', '') if validPatternAttr is not None else None
                if added_pattern is not None:
                    validPattern = added_pattern

                if my_type == 'Edm.String' and enum_annotation is not None:
                    memberList = enum_annotation.find('Collection').find_all('PropertyValue', attrs={'Property': 'Member'})
                    validPattern = '|'.join([re.escape(x.get('String')) for x in memberList if x.get('String')])

                return RedfishProperty.validate_basic(val, my_type, validPattern, validMin, validMax)
        return True
    
    def as_json(self):
        return self.createObject().as_json()
    
    def createObject(self, name='Object'):
        return RedfishObject(self, name)
                

class RedfishProperty(object):
    """Property in a resource
    Represents all Types given, however, ComplexTypes are better suited to be RedfishObjects
    """
    def __repr__(self):
        if self.Populated:
            return "{}--{}, Value: {}".format(self.Name, self.Type, self.Value)
        else:
            return "{}--{}".format(self.Name, self.Type)

    def __init__(self, my_type, name="Property", parent=None):
        self.Name = name
        self.Type = my_type
        self.HasSchema = self.Type != REDFISH_ABSENT
        self.Populated = False
        self.Value = None
        self.IsValid = False
        self.IsCollection = False
        self.InAnnotation = False
        self.IsAutoExpanded = False
        self.IsExcerpt = False
        self.SchemaExists = False
        self.Exists = False
        self.parent = parent
        self.added_pattern = None

    def populate(self, val, check=False):
        eval_prop = copy.copy(self)
        eval_prop.Populated = True
        eval_prop.Value = val
        eval_prop.IsCollection = isinstance(val, list)
        eval_prop.InAnnotation = False
        eval_prop.SchemaExists = True
        eval_prop.Exists = val != REDFISH_ABSENT

        eval_prop.IsValid = True

        if isinstance(eval_prop.Type, str):
            is_type_collection = 'Collection(' in eval_prop.Type
        elif isinstance(eval_prop.Type, RedfishType):
            is_type_collection = eval_prop.Type.IsCollection()
        else:
            raise ValueError('Type is not String or RedfishType')
        if eval_prop.IsCollection and not is_type_collection:
            my_logger.error('Property {} should not be a List'.format(self.Name))
            eval_prop.IsValid = False
        elif not eval_prop.IsCollection and is_type_collection and val not in [None, REDFISH_ABSENT]:
            my_logger.error('Collection Property {} is not a List'.format(self.Name))
            eval_prop.IsValid = False

        if isinstance(eval_prop.Type, str) and 'Edm.' in eval_prop.Type and check:
            try:
                eval_prop.IsValid = eval_prop.Exists and RedfishProperty.validate_basic(eval_prop.Value, eval_prop.Type)
            except ValueError as e:
                my_logger.error('{}: {}'.format(eval_prop.Name, e))  # log this
                eval_prop.IsValid = False
        elif isinstance(eval_prop.Type, RedfishType) and check:
            try:
                eval_prop.IsValid = eval_prop.Type.validate(eval_prop.Value, eval_prop.added_pattern)
            except ValueError as e:
                my_logger.error('{}: {}'.format(eval_prop.Name, e))  # log this
                eval_prop.IsValid = False
        return eval_prop

    def as_json(self):
        my_dict = {x: y for x, y in vars(self).items() if x in ['Name', 'Type', 'Value', 'IsValid', 'Exists', 'SchemaExists']}
        if isinstance(self.Type, RedfishType):
            my_dict['IsRequired'] = self.Type.IsMandatory
            my_dict['IsNullable'] = self.Type.IsNullable
            my_dict['HasAdditional'] = self.Type.HasAdditional
        return my_dict
    
    def getLinks(self):
        return []
    
    @staticmethod
    def validate_string(val, pattern):
        """validateString

        Validates a string, given a value and a pattern
        """
        if not isinstance(val, str):
            raise ValueError(
                "Expected string value, got type {}".format(str(type(val)).strip("<>"))
            )
        if pattern is not None:
            match = re.fullmatch(pattern, val)
            if match is None:
                raise ValueError(
                    "String '{}' does not match pattern '{}'".format(
                        str(val), repr(pattern)
                    )
                )
        return True

    @staticmethod
    def validate_number(val, minVal=None, maxVal=None):
        """validateNumber

        Validates a Number and its min/max values
        """

        if not isinstance(val, (int, float)):
            raise ValueError(
                "Expected integer or float, got type {}".format(
                    str(type(val)).strip("<>")))
        if minVal is not None:
            if not minVal <= val:
                raise ValueError(
                    "Value out of assigned min range, {} > {}".format(
                        str(val), str(minVal)))
        if maxVal is not None:
            if not maxVal >= val:
                raise ValueError(
                    "Value out of assigned max range, {} > {}".format(
                        str(val), str(maxVal)))
        return True

    @staticmethod
    def validate_basic(val, my_type, validPattern=None, min=None, max=None):
        if "Collection(" in my_type:
            my_type = my_type.replace("Collection(", "").replace(")", "")

        if isinstance(val, list):
            return all([RedfishProperty.validate_basic(sub_val, my_type, validPattern, min, max) for sub_val in val])

        elif my_type == "Edm.Boolean":
            if not isinstance(val, bool):
                raise ValueError(
                    "Expected bool, got type {}".format(str(type(val)).strip("<>")))
            return True

        elif my_type == "Edm.DateTimeOffset":
            return RedfishProperty.validate_string(
                val, r".*(Z|(\+|-)[0-9][0-9]:[0-9][0-9])")

        elif my_type == "Edm.Duration":
            return RedfishProperty.validate_string(
                val, r"P([0-9]+D)?(T([0-9]+H)?([0-9]+M)?([0-9]+(\.[0-9]+)?S)?)?")

        elif my_type == "Edm.Guid":
            return RedfishProperty.validate_string(
                val, r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}")

        elif my_type == "Edm.String":
            return RedfishProperty.validate_string(val, validPattern)

        elif my_type in ["Edm.Int16", "Edm.Int32", "Edm.Int64", "Edm.Int"]:
            if not isinstance(val, int):
                raise ValueError("Expected int, got type {}".format(str(type(val)).strip("<>")))
            return RedfishProperty.validate_number(val, min, max)

        elif my_type == "Edm.Decimal" or my_type == "Edm.Double":
            return RedfishProperty.validate_number(val, min, max)

        elif my_type == "Edm.Primitive" or my_type == "Edm.PrimitiveType":
            if not isinstance(val, (int, float, str, bool)):
                raise ValueError("Expected primitive, got type {}".format(str(type(val)).strip("<>")))
            return True

        else:
            return False

class RedfishObject(RedfishProperty):
    """Represents Redfish as they are represented as Resource/ComplexTypes

    Can be indexed with []
    Can be populated with Data
    If Populated, can be grabbed for Links
    Can get json representation of type properties with as_json
    """
    def __getitem__(self, index):
        return self.properties[index]

    def __contains__(self, item):
        if self.Populated:
            return item in self.properties and self.properties[item].Exists
        else:
            return item in self.properties

    def __init__(self, redfish_type: RedfishType, name="Object", parent=None):
        super().__init__(redfish_type, name, parent)
        self.payload = None
        self.HasValidUri = False
        self.HasValidUriStrict = False
        self.properties = {}
        for prop, typ in redfish_type.getProperties().items():
            try:
                base = typ.getBaseType()
                if base == 'complex':
                    self.properties[prop] = RedfishObject(typ, prop, self)
                else:
                    self.properties[prop] = RedfishProperty(typ, prop, self)
            except MissingSchemaError:
                self.properties[prop] = RedfishProperty(REDFISH_ABSENT, prop, self)
                my_logger.warning('Schema not found for {}'.format(typ))

    def populate(self, payload, check=False, casted=False):
        """
        Return a populated object, or list of objects
        """
        populated_object = super().populate(payload)
        populated_object.payload = payload

        if isinstance(payload, list):
            populated_object.IsValid = populated_object.Type.IsCollection()
            if not populated_object.IsValid:
                my_logger.error("This object's type {} should be a Collection, but it's of type {}...".format(populated_object.Name, populated_object.Type))
                return populated_object
            populated_object.IsCollection = True

            my_new_type = stripCollection(populated_object.Type.fulltype)

            new_type_obj = populated_object.Type.catalog.getSchemaDocByClass(getNamespace(my_new_type)).getTypeInSchemaDoc(my_new_type)

            new_rf_object = RedfishObject(new_type_obj, populated_object.Name, populated_object.parent)

            populated_object.Value = [new_rf_object.populate(sub_item, check, casted) for sub_item in payload]
            return populated_object
        else:
            if populated_object.Type.IsCollection():
                if payload in [REDFISH_ABSENT, None]:
                    populated_object.Value = payload
                    return populated_object
                else:
                    my_logger.error("This object {} should be a list, but it's of type {}...".format(populated_object.Name, type(payload).__name__))
                    return populated_object

        # todo: redesign objects to have consistent variables, not only when populated
        # if populated, should probably just use another python class?
        # remember that populated RedfishObjects may have more objects embedded in them
        # i.e. OEM or complex or arrays
        if payload == REDFISH_ABSENT or payload is None:
            populated_object.HasValidUri = True
            populated_object.HasValidUriStrict = False
            populated_object.properties = {x: y.populate(REDFISH_ABSENT) for x, y in populated_object.properties.items()}
            return populated_object

        populated_object.IsValid = isinstance(payload, dict)

        # Only valid if we are a dictionary...
        # todo: see above None/REDFISH_ABSENT block
        if not populated_object.IsValid:
            my_logger.error("This complex object {} should be a dictionary or None, but it's of type {}...".format(populated_object.Name, type(payload).__name__))
            populated_object.HasValidUri = True
            populated_object.HasValidUriStrict = False
            populated_object.properties = {x: y.populate(REDFISH_ABSENT) for x, y in populated_object.properties.items()}

        # Cast types if they're below their parent or are OemObjects
        # Don't cast objects with odata.type that matches their current object type
        already_typed = False
        sub_payload = payload
        my_odata_type = sub_payload.get('@odata.type')
        if my_odata_type is not None and str(populated_object.Type) == my_odata_type.strip('#'):
            already_typed = True
        if populated_object.Type.IsNav:
            already_typed = True

        # If our item is an OemObject type and hasn't been casted to a type, then cast it
        if 'Resource.OemObject' in populated_object.Type.getTypeTree() and not casted:
            my_logger.verbose1(('Morphing OemObject', my_odata_type, populated_object.Type))
            if my_odata_type:
                my_odata_type = my_odata_type.strip('#')
                try:
                    type_obj = populated_object.Type.catalog.getSchemaDocByClass(my_odata_type).getTypeInSchemaDoc(my_odata_type)
                    populated_object = RedfishObject(type_obj, populated_object.Name, populated_object.parent).populate(sub_payload, check=check, casted=True)
                except MissingSchemaError:
                    my_logger.warning("Couldn't get schema for object, skipping OemObject {}".format(populated_object.Name))
                except Exception as e:
                    my_logger.warning("Couldn't get schema for object (?), skipping OemObject {} : {}".format(populated_object.Name, e))
                return populated_object
        # Otherwise, if we're not casted, or we don't have an odata type, then cast it
        elif not casted and not already_typed:
            my_ns, my_ns_unversioned = populated_object.Type.Namespace, getNamespaceUnversioned(populated_object.Type.Namespace)
            # if we have an odata type, use it as our upper limit
            if my_odata_type:
                my_limit = getNamespace(my_odata_type).strip('#')
            else:
                my_limit = 'v9_9_9'
            # If our item is not a Resource.Resource type, determine its parent's version limit for later...
            # NOTE: Resource items always seem to be cast to highest type, not determined by its parent's type
            #       not sure where this is backed up in documentation
            if populated_object.parent and my_ns_unversioned not in ['Resource']:
                parent = populated_object
                while parent.parent and parent.parent.Type.Namespace.startswith(my_ns_unversioned + '.'):
                    parent = parent.parent
                    my_limit = parent.Type.Namespace

            my_type = populated_object.Type.TypeName
            top_ns = my_ns

            # get type order from bottom up of SchemaDoc
            for top_ns, schema in reversed(list(populated_object.Type.catalog.getSchemaDocByClass(my_ns).classes.items())):
                # if our object type is in schema... check for limit
                if my_type in schema.my_types:
                    if splitVersionString(top_ns) <= splitVersionString(my_limit):
                        my_ns = top_ns
                        break
            # ISSUE: We can't cast under v1_0_0, get the next best Type
            if my_ns == my_ns_unversioned:
                my_ns = top_ns
            if my_ns not in populated_object.Type.Namespace:
                # NOTE: This returns a Type object without IsPropertyType
                my_logger.verbose1(('Morphing Complex', my_ns, my_type, my_limit))
                new_type_obj = populated_object.Type.catalog.getSchemaDocByClass(my_ns).getTypeInSchemaDoc('.'.join([my_ns, my_type]))
                populated_object = RedfishObject(new_type_obj, populated_object.Name, populated_object.parent).populate(sub_payload, check=check, casted=True)
                return populated_object

        # Validate our Uri
        populated_object.HasValidUri = True
        populated_object.HasValidUriStrict = True
        allowable_uris = populated_object.Type.getUris()
        # If we have expected URIs and @odata.id
        # And we AREN'T a navigation property
        if not populated_object.Type.catalog.flags['ignore_uri_checks'] and len(allowable_uris) and '@odata.id' in sub_payload:
            # Strip our URI and warn if that's the case
            my_odata_id = sub_payload['@odata.id']
            if my_odata_id != '/redfish/v1/' and my_odata_id.endswith('/'):
                if check: my_logger.warning('Stripping end of URI... {}'.format(my_odata_id))
                my_odata_id = my_odata_id.rstrip('/')

            # Initial check if our URI matches our format at all
            # Setup REGEX...
            my_uri_regex = "^{}$".format("|".join(allowable_uris))
            my_uri_regex = re.sub(URI_ID_REGEX, VALID_ID_REGEX, my_uri_regex)
            populated_object.HasValidUri = re.fullmatch(my_uri_regex, my_odata_id) is not None
            populated_object.HasValidUriStrict = populated_object.HasValidUri

            if 'Resource.Resource' in populated_object.Type.getTypeTree():
                if '#' in my_odata_id:
                    my_logger.warning('Found uri with fragment, which Resource.Resource types do not use {}'.format(my_odata_id))
            elif 'Resource.ReferenceableMember' in populated_object.Type.getTypeTree():
                if '#' not in my_odata_id:
                    my_logger.warning('No fragment in URI, but ReferenceableMembers require it {}'.format(my_odata_id))

            # check that our ID is matching
            # this won't check NavigationProperties but the Resources will
            if populated_object.HasValidUri and not populated_object.Type.IsNav:
                # pair our type, Id value, and odata.id value
                my_odata_split = my_odata_id.split('/')
                my_type, my_id, my_uri_id = populated_object.Type.TypeName, sub_payload.get('Id'), my_odata_split[-1]

                for schema_uri in allowable_uris:
                    # regex URI check to confirm which URI 
                    my_uri_regex = re.sub(URI_ID_REGEX, VALID_ID_REGEX, "^{}$".format(schema_uri))
                    if re.fullmatch(my_uri_regex, my_odata_id):
                        # pair our uri with the current resource
                        schema_uri_end = schema_uri.rsplit('/')[-1]
                        # if our Uri is expecting an Id, then check if they match, otherwise we are already passing
                        if re.match(URI_ID_REGEX, schema_uri_end):
                            if my_id is not None:
                                populated_object.HasValidUriStrict = my_id == my_uri_id
                        break

        # TODO: Oem support is able, but it is tempermental for Actions and Additional properties
        #if 'Resource.OemObject' in sub_obj.Type.getTypeTree():
        #    evals.append(sub_obj)
        #    continue

        # populate properties
        if populated_object.Name == 'Actions':
            populated_object.properties = {x: y.populate(sub_payload.get(x, REDFISH_ABSENT)) for x, y in populated_object.properties.items() if x != 'Oem'}
        else:
            populated_object.properties = {x: y.populate(sub_payload.get(x, REDFISH_ABSENT)) for x, y in populated_object.properties.items()}

        # additional_props
        if populated_object.Type.DynamicProperties:
            my_dynamic = populated_object.Type.DynamicProperties
            my_odata_type = my_dynamic.get('Type', 'Resource.OemObject')
            prop_pattern = my_dynamic.get('Pattern', '.*')
            allow_property_generation = populated_object.Name != 'Actions'
        else:
            my_odata_type = 'Resource.OemObject'
            prop_pattern = '.*'
            allow_property_generation = populated_object.Type.HasAdditional and populated_object.Name != 'Actions'

        if allow_property_generation:
            my_property_names = [x for x in sub_payload if x not in populated_object.properties if re.match(prop_pattern, x) and '@' not in x]
            for add_name in my_property_names:
                if 'Edm.' in my_odata_type:
                    my_new_term = '<Term Name="{}" Type="{}"> </Term>'.format(add_name, my_odata_type) # Make a pseudo tag because RedfishType requires it...
                    new_soup = BeautifulSoup(my_new_term, "xml").find('Term')
                    type_obj = RedfishType(new_soup, populated_object.Type.owner)
                else:
                    type_obj = populated_object.Type.catalog.getSchemaDocByClass(my_odata_type).getTypeInSchemaDoc(my_odata_type)
                if type_obj.getBaseType() == 'complex':
                    object = RedfishObject(type_obj, name=add_name, parent=populated_object)
                else:
                    object = RedfishProperty(type_obj, name=add_name, parent=populated_object)
                my_logger.debug('Populated {} with {}'.format(my_property_names, object.as_json()))
                my_logger.verbose1(('Adding Additional', add_name, my_odata_type, populated_object.Type))
                populated_object.properties[add_name] = object.populate(sub_payload.get(add_name, REDFISH_ABSENT))

        my_annotations = [x for x in sub_payload if x not in populated_object.properties if '@' in x and '@odata' not in x]
        for key in my_annotations:
            splitKey = key.split('@', 1)
            fullItem = splitKey[1]
            if getNamespace(fullItem) not in allowed_annotations:
                my_logger.warning("getAnnotations: {} is not an allowed annotation namespace, please check spelling/capitalization.".format(fullItem))
                continue
            try:
                type_obj = populated_object.Type.catalog.getSchemaInCatalog(fullItem).terms[getType(fullItem)]
                if type_obj.getBaseType() == 'complex':
                    object = RedfishObject(type_obj, name=key, parent=populated_object)
                else:
                    object = RedfishProperty(type_obj, name=key, parent=self)
                my_logger.verbose1(('Adding Additional', key, my_odata_type, populated_object.Type))
                populated_object.properties[key] = object.populate(sub_payload[key])
            except:
                my_logger.error("Unable to locate the definition of the annotation '@{}'.".format(fullItem))

        return populated_object

    def as_json(self):
        if self.Populated:
            return {'Properties' : {a: b.as_json() for a, b in self.properties.items() if (b.Exists or not b.IsValid)}}
        else:
            # base = {'Self': super().as_json()}
            base = super().as_json()
            base.update({'Properties': {a: b.as_json() for a, b in self.properties.items()}})
            return base

    def getLinks(self):
        """Grab links from our Object
        """
        links = []
        # if we're populated...
        if self.Populated:
            for n, property in self.properties.items():
                # if we don't exist or our type is Basic
                if not isinstance(property, list):
                    property = [property]
                for item in property:
                    if not item.Exists: continue
                    if not isinstance(item.Type, RedfishType): continue
                    if n == 'Actions':
                        new_type = item.Type.catalog.getTypeInCatalog('ActionInfo.ActionInfo')
                        for act in item.Value.values():
                            if isinstance(act, dict):
                                uri = act.get('@Redfish.ActionInfo')
                                if isinstance(uri, str):
                                    my_link = RedfishObject(new_type, 'ActionInfo', item).populate({'@odata.id': uri})
                                    my_link.InAnnotation = True
                                    links.append(my_link)
                    if item.Type.IsNav:
                        if isinstance(item.Value, list):
                            for num, val in enumerate(item.Value):
                                # TODO: Along with example Excerpt and RedfishObject, replace following code with hypothetical RedfishType.getCollectionType
                                try:
                                    new_type_obj = item.Type.getCollectionType()
                                    new_link = RedfishObject(new_type_obj, item.Name, item.parent).populate(val)
                                    new_link.Name = new_link.Name + '#{}'.format(num)
                                    if item.Type.AutoExpand:
                                        new_link.IsAutoExpanded = True
                                    if item.Type.Excerpt:
                                        new_link.IsExcerpt = True
                                    links.append(new_link)
                                except Exception as e:
                                    my_logger.error('Unable to build definition for URI {}; check its schema definition or the schema making the reference to the URI for schema errors: {}'.format(val, repr(e)))
                        else:
                            links.append(item)
                    elif item.Type.getBaseType() == 'complex':
                        if item.Value is None:
                            continue
                        InAnnotation = item.Name in ['@Redfish.Settings', '@Redfish.ActionInfo', '@Redfish.CollectionCapabilities']
                        my_links = item.getLinks()
                        for sub_item in my_links:
                            sub_item.InAnnotation = InAnnotation
                        links.extend(my_links)
        return links



# Copyright Notice:
# Copyright 2016-2020 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

import re
import logging
from types import SimpleNamespace

# TODO: Replace logger with custom logger with custom verbose levels, remove verbose1 and verbose2 
my_logger = logging.getLogger()
my_logger.setLevel(logging.DEBUG)

"""
 Power.1.1.1.Power , Power.v1_0_0.Power
"""

versionpattern = 'v[0-9]+_[0-9]+_[0-9]+'

LOG_ENTRY = ('name', 'value', 'type', 'exists', 'result')

def create_entry(name, value, type, exists, result):
    return SimpleNamespace(**{
        "name": name,
        "value": value,
        "type": type,
        "exists": exists,
        "result": result
    })


def splitVersionString(v_string):
    """
    Split x.y.z and Namespace.vX_Y_Z, vX_Y_Z type version strings into tuples of integers

    :return: tuple of integers
    """
    if(re.match('([a-zA-Z0-9_.-]*\.)+[a-zA-Z0-9_.-]*', v_string) is not None):
        new_string = getVersion(v_string)
        if new_string is not None:
            v_string = new_string
    if ('_' in v_string):
        v_string = v_string.replace('v', '')
        payload_split = v_string.split('_')
    else:
        payload_split = v_string.split('.')
    if len(payload_split) != 3:
        return tuple([0, 0, 0])
    return tuple([int(v) for v in payload_split])


def stripCollection(typename):
    """
    Remove "Collection()" from a type string
    """
    if 'Collection(' in typename:
        typename = typename.replace('Collection(', "").replace(')', "")
    return typename


def navigateJsonFragment(decoded, URILink):
    if '#' in URILink:
        URIfragless, frag = tuple(URILink.rsplit('#', 1))
        fragNavigate = frag.split('/')
        for item in fragNavigate:
            if item == '':
                continue
            if isinstance(decoded, dict):
                decoded = decoded.get(item)
            elif isinstance(decoded, list):
                if not item.isdigit():
                    my_logger.error("This URI ({}) is accessing an array, but this is not an index: {}".format(URILink, item))
                    return None
                if int(item) >= len(decoded):
                    my_logger.error("This URI ({}) is accessing an array, but the index is too large for an array of size {}: {}".format(URILink, len(decoded), item))
                    return None
                decoded = decoded[int(item)]
            else:
                my_logger.error("This URI ({}) has resolved to an invalid object that is neither an array or dictionary".format(URILink))
                return None
    return decoded


def getNamespace(string: str):
    """getNamespace

    Gives namespace of a type string, version included

    :param string:  A type string
    :type string: str
    """
    if '#' in string:
        string = string.rsplit('#', 1)[1]
    return string.rsplit('.', 1)[0]


def getVersion(string: str):
    """getVersion

    Gives version stripped from type/namespace string, if possible

    :param string:  A type/namespace string
    :type string: str
    """
    regcap = re.search(versionpattern, string)
    return regcap.group() if regcap else None


def getNamespaceUnversioned(string: str):
    """getNamespaceUnversioned

    Gives namespace of a type string, version NOT included

    :param string:
    :type string: str
    """
    if '#' in string:
        string = string.rsplit('#', 1)[1]
    return string.split('.', 1)[0]


def getType(string: str):
    """getType

    Gives type of a type string (right hand side)

    :param string:
    :type string: str
    """
    if '#' in string:
        string = string.rsplit('#', 1)[1]
    return string.rsplit('.', 1)[-1]


def createContext(typestring: str):
    """createContext

    Create an @odata.context string from a type string

    :param typestring:
    :type string: str
    """
    ns_name = getNamespaceUnversioned(typestring)
    type_name = getType(typestring)
    context = '/redfish/v1/$metadata' + '#' + ns_name + '.' + type_name
    return context


def checkPayloadConformance(jsondata, uri):
    """
    checks for @odata entries and their conformance
    These are not checked in the normal loop
    """
    info = {}
    decoded = jsondata
    success = True
    for key in [k for k in decoded if '@odata' in k]:
        paramPass = False

        property_name, odata_name = tuple(key.rsplit('@', maxsplit=1))

        if odata_name == 'odata.id':
            paramPass = isinstance(decoded[key], str)
            paramPass = re.match(
                '(\/.*)+(#([a-zA-Z0-9_.-]*\.)+[a-zA-Z0-9_.-]*)?', decoded[key]) is not None
            if not paramPass:
                my_logger.error("{} {}: Expected format is /path/to/uri, but received: {}".format(uri, key, decoded[key]))
            else:
                if uri != '' and decoded[key] != uri and not (uri == "/redfish/v1/" and decoded[key] == "/redfish/v1"):
                    my_logger.warning("{} {}: Expected @odata.id to match URI link {}".format(uri, key, decoded[key]))
        elif odata_name == 'odata.count':
            paramPass = isinstance(decoded[key], int)
            if not paramPass:
                my_logger.error("{} {}: Expected an integer, but received: {}".format(uri, key, decoded[key]))
        elif odata_name == 'odata.context':
            paramPass = isinstance(decoded[key], str)
            paramPass = re.match(
                '/redfish/v1/\$metadata#([a-zA-Z0-9_.-]*\.)[a-zA-Z0-9_.-]*', decoded[key]) is not None
            if not paramPass:
                my_logger.warning("{} {}: Expected format is /redfish/v1/$metadata#ResourceType, but received: {}".format(uri, key, decoded[key]))
                info[key] = (decoded[key], 'odata', 'Exists', 'WARN')
                continue
        elif odata_name == 'odata.type':
            paramPass = isinstance(decoded[key], str)
            paramPass = re.match(
                '#([a-zA-Z0-9_.-]*\.)+[a-zA-Z0-9_.-]*', decoded[key]) is not None
            if not paramPass:
                my_logger.error("{} {}: Expected format is #Namespace.Type, but received: {}".format(uri, key, decoded[key]))
        else:
            paramPass = True

        success = success and paramPass

        info[key] = (decoded[key], 'odata', 'Exists', 'PASS' if paramPass else 'FAIL')
        
    return success, info



# Copyright Notice:
# Copyright 2016-2021 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

import configparser
import logging
import json

my_logger = logging.getLogger()
my_logger.setLevel(logging.DEBUG)

config_struct = {
    'Tool': ['verbose'],
    'Host': ['ip', 'username', 'password', 'description', 'forceauth', 'authtype', 'token', 'ext_http_proxy', 'ext_https_proxy', 'serv_http_proxy', 'serv_https_proxy'],
    'Validator': ['payload', 'logdir', 'oemcheck', 'debugging', 'schema_directory', 'uricheck', 'mockup']
}

config_options = [x for name in config_struct for x in config_struct[name]]


def convert_args_to_config(args):
    # Disable interpolation (https://docs.python.org/3/library/configparser.html#interpolation-of-values)
    my_config = configparser.ConfigParser(interpolation=None)
    for section in ['Tool', 'Host', 'Validator']:
        my_config.add_section(section)
        for option in config_struct[section]:
            if option not in ['password', 'token']:
                my_var = vars(args)[option]
                if isinstance(my_var, list):
                    my_var = ' '.join(my_var)
                    print(my_var)
                my_config.set(section, option, str(my_var) if my_var is not None else '')
            else:
                my_config.set(section, option, '******')
    return my_config


def convert_config_to_args(args, config):
    my_config = configparser.ConfigParser()
    if isinstance(config, configparser.ConfigParser):
        my_config = config
    elif isinstance(config, str):
        with open(config, 'r') as f:
            my_config.read_file(f)
    elif isinstance(config, dict):
        my_config.read_dict(config)
    for section in config_struct:
        if section in my_config:
            for option in my_config[section]:
                if option.lower() not in config_options:
                    if option.lower() not in ['version', 'copyright']:
                        my_logger.error('Option {} not supported!'.format(option))
                elif my_config[section][option] not in ['', None]:
                    if option.lower() == 'payload':
                        setattr(args, option, my_config[section][option].split(' '))
                    else:
                        setattr(args, option, my_config[section][option])
    my_config_dict = config_parse_to_dict(my_config)
    print(json.dumps(my_config_dict, indent=4))
        

def config_parse_to_dict(config):
    my_dict = {}
    for section in config:
        my_dict[section] = {}
        for option in [x for x in config[section] if x not in ['version', 'copyright']]:
            my_dict[section][option] = {}
            my_dict[section][option]['value'] = config[section][option]
            my_dict[section][option]['description'] = "TBD"
    return my_dict



# Copyright Notice:
# Copyright 2016-2019 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

if __name__ != '__main__':
    from redfish_service_validator.helper import getType
else:
    import argparse
    from bs4 import BeautifulSoup
    import os, csv
import redfish_service_validator.RedfishLogo as logo
from redfish_service_validator.helper import LOG_ENTRY
from types import SimpleNamespace
from collections import Counter
import json
import re

# hack in tagnames into module namespace
tag = SimpleNamespace(**{tagName: lambda string, attr=None, tag=tagName: wrapTag(string, tag=tag, attr=attr)\
    for tagName in ['tr', 'td', 'th', 'div', 'b', 'table', 'body', 'head', 'summary']})

def count_errors(results):
    error_lines = []
    finalCounts = Counter()
    for item in results:
        innerCounts = results[item]['counts']

        # detect if there are error messages for this resource, but no failure counts; if so, add one to the innerCounts
        counters_all_pass = True
        for countType in sorted(innerCounts.keys()):
            if innerCounts.get(countType) == 0:
                continue
            if re.search(r"^error|problem|fail|bad|exception|err[.]", countType):
                counters_all_pass = False
                error_lines.append('{} {} errors in {}'.format(innerCounts[countType], countType, results[item]['uri']))
            innerCounts[countType] += 0
        error_messages_present = False
        if results[item]['errors'] is not None and len(results[item]['errors']) > 0:
            error_messages_present = True
        if results[item]['warns'] is not None and len(results[item]['warns']) > 0:
            innerCounts['warningPresent'] = 1
        if counters_all_pass and error_messages_present:
            innerCounts['failErrorPresent'] = 1
            error_lines.append('Error message present in {}'.format(results[item]['uri']))

        finalCounts.update(innerCounts)
    return error_lines, finalCounts

def wrapTag(string, tag='div', attr=None):
    string = str(string)
    ltag, rtag = '<{}>'.format(tag), '</{}>'.format(tag)
    if attr is not None:
        ltag = '<{} {}>'.format(tag, attr)
    return ltag + string + rtag


def infoBlock(strings, split='<br/>', ffunc=None, sort=True):
    if isinstance(strings, dict):
        infos = [tag.b('{}: '.format(y)) + str(x) for y, x in (sorted(strings.items()) if sort else strings.items())]
    else:
        infos = strings
    return split.join([ffunc(*x) for x in enumerate(infos)] if ffunc is not None else infos)


def tableBlock(lines, titles, widths=None, ffunc=None):
    widths = widths if widths is not None else [100 for x in range(len(titles))]
    attrlist = ['style="width:{}%"'.format(str(x)) for x in widths]
    tableHeader = tag.tr(''.join([tag.th(x, y) for x, y in zip(titles,attrlist)]))
    for line in lines:
        tableHeader += tag.tr(''.join([ffunc(cnt, x) if ffunc is not None else tag.td(x) for cnt, x in enumerate(line)]))
    return tag.table(tableHeader)


def applySuccessColor(num, entry):
    if num < 4:
        return wrapTag(entry, 'td')
    success_col = str(entry)
    if any(x.upper() in str(success_col).upper() for x in ['FAIL', 'errorExcerpt']):
        entry = '<td class="fail center">' + str(success_col) + '</td>'
    elif str(success_col).upper() in ['DEPRECATED', 'INVALID', 'WARN']:
        entry = '<td class="warn center">' + str(success_col) + '</td>'
    elif any(x in str(success_col).upper() for x in ['DEPRECATED', 'INVALID', 'WARN']):
        entry = '<td class="warn center">' + str(success_col) + '</td>'
    elif 'PASS' in str(success_col).upper():
        entry = '<td class="pass center">' + str(success_col) + '</td>'
    else:
        entry = '<td class="center">' + str(success_col) + '</td>'
    return entry


def applyInfoSuccessColor(num, entry):
    if any(x in entry for x in ['fail', 'exception', 'error', 'problem', 'err']):
        style = 'class="fail"'
    elif 'warn' in entry or 'invalid' in entry:
        style = 'class="warn"'
    else:
        style = None
    return tag.div(entry, attr=style)


def renderHtml(results, tool_version, startTick, nowTick, service):
    # Render html
    config = service.config
    sysDescription, ConfigURI = (config['description'], config['ip'])
    error_lines, finalCounts = count_errors(results)
    if service.metadata is not None:
        finalCounts.update(service.metadata.get_counter())

    # wrap html
    htmlPage = ''
    htmlStrTop = '<head><title>Conformance Test Summary</title>\
            <style>\
            .pass {background-color:#99EE99}\
            .column {\
                float: left;\
                width: 45%;\
            }\
            .fail {background-color:#EE9999}\
            .warn {background-color:#EEEE99}\
            .bluebg {background-color:#BDD6EE}\
            .button {padding: 12px; display: inline-block}\
            .center {text-align:center;}\
            .log {text-align:left; white-space:pre-wrap; word-wrap:break-word; font-size:smaller; padding: 6px}\
            .title {background-color:#DDDDDD; border: 1pt solid; font-height: 30px; padding: 8px}\
            .titlesub {padding: 8px}\
            .titlerow {border: 2pt solid}\
            .results {transition: visibility 0s, opacity 0.5s linear; display: none; opacity: 0}\
            .resultsShow {display: block; opacity: 1}\
            body {background-color:lightgrey; border: 1pt solid; text-align:center; margin-left:auto; margin-right:auto}\
            th {text-align:center; background-color:beige; border: 1pt solid}\
            td {text-align:left; background-color:white; border: 1pt solid; word-wrap:break-word; overflow:hidden;}\
            table {width:90%; margin: 0px auto; table-layout:fixed;}\
            .titletable {width:100%}\
            </style>\
            </head>'
    htmlStrBodyHeader = ''
    # Logo and logname
    infos = [wrapTag('##### Redfish Conformance Test Report #####', 'h2')]
    infos.append(wrapTag('<img align="center" alt="DMTF Redfish Logo" height="203" width="288"'
                         'src="data:image/gif;base64,' + logo.logo + '">', 'h4'))
    infos.append('<h4><a href="https://github.com/DMTF/Redfish-Service-Validator">'
                 'https://github.com/DMTF/Redfish-Service-Validator</a></h4>')
    infos.append('Tool Version: {}'.format(tool_version))
    infos.append(startTick.strftime('%c'))
    infos.append('(Run time: {})'.format(
        str(nowTick - startTick).rsplit('.', 1)[0]))
    infos.append('<h4>This tool is provided and maintained by the DMTF. '
                 'For feedback, please open issues<br>in the tool\'s Github repository: '
                 '<a href="https://github.com/DMTF/Redfish-Service-Validator/issues">'
                 'https://github.com/DMTF/Redfish-Service-Validator/issues</a></h4>')

    htmlStrBodyHeader += tag.tr(tag.th(infoBlock(infos)))

    infos = {x: config[x] for x in config if x not in ['systeminfo', 'ip', 'password', 'description']}
    infos_left, infos_right = dict(), dict()
    for key in sorted(infos.keys()):
        if len(infos_left) <= len(infos_right):
            infos_left[key] = infos[key]
        else:
            infos_right[key] = infos[key]

    block = tag.td(tag.div(infoBlock(infos_left), 'class=\'column log\'') \
            + tag.div(infoBlock(infos_right), 'class=\'column log\''), 'id=\'resNumConfig\' class=\'results resultsShow\'')

    htmlButtons = '<div class="button warn" onClick="arr = document.getElementsByClassName(\'results\'); for (var i = 0; i < arr.length; i++){arr[i].classList.add(\'resultsShow\')};">Expand All</div>'
    htmlButtons += '<div class="button fail" onClick="arr = document.getElementsByClassName(\'results\'); for (var i = 0; i < arr.length; i++){arr[i].classList.remove(\'resultsShow\')};">Collapse All</div>'
    htmlButtons += tag.div('Toggle Config', attr='class="button pass" onClick="document.getElementById(\'resNumConfig\').classList.toggle(\'resultsShow\');"')

    htmlStrBodyHeader += tag.tr(tag.th(htmlButtons))
    htmlStrBodyHeader += tag.tr(tag.th('Test Summary'))

    infos = {'System': ConfigURI, 'Description': sysDescription}
    htmlStrBodyHeader += tag.tr(tag.th(infoBlock(infos)))

    errors = error_lines
    if len(errors) == 0:
        errors = ['No errors located.']
    errorTags = tag.td(infoBlock(errors), 'class="log"')

    infos_left, infos_right = dict(), dict()
    for key in sorted(finalCounts.keys()):
        if finalCounts.get(key) == 0:
            continue
        if len(infos_left) <= len(infos_right):
            infos_left[key] = finalCounts[key]
        else:
            infos_right[key] = finalCounts[key]

    htmlStrCounts = (tag.div(infoBlock(infos_left), 'class=\'column log\'') + tag.div(infoBlock(infos_right), 'class=\'column log\''))

    htmlStrBodyHeader += tag.tr(tag.td(htmlStrCounts))

    htmlStrBodyHeader += tag.tr(errorTags)

    htmlStrBodyHeader += tag.tr(block)


    if service.metadata is not None:
        htmlPage = service.metadata.to_html()
    for cnt, item in enumerate(results):
        entry = []
        val = results[item]
        rtime = '(response time: {})'.format(val['rtime'])
        rcode = '({})'.format(val['rcode'])
        payload = val.get('payload', {})

        # uri block
        prop_type = val['fulltype']
        if prop_type is not None:
            type_name = getType(prop_type)

        infos = [str(val.get(x)) for x in ['uri', 'samplemapped'] if val.get(x) not in ['',None]]
        infos.append(rtime)
        infos.append(type_name)
        uriTag = tag.tr(tag.th(infoBlock(infos, '&ensp;'), 'class="titlerow bluebg"'))
        entry.append(uriTag)

        # info block
        infos = [str(val.get(x)) for x in ['uri'] if val.get(x) not in ['',None]]
        infos.append(rtime)
        infos.append(tag.div('Show Results', attr='class="button warn"\
                onClick="document.getElementById(\'payload{}\').classList.remove(\'resultsShow\');\
                document.getElementById(\'resNum{}\').classList.toggle(\'resultsShow\');"'.format(cnt, cnt)))
        infos.append(tag.div('Show Payload', attr='class="button pass"\
                onClick="document.getElementById(\'payload{}\').classList.toggle(\'resultsShow\');\
                document.getElementById(\'resNum{}\').classList.add(\'resultsShow\');"'.format(cnt, cnt)))
        buttonTag = tag.td(infoBlock(infos), 'class="title" style="width:30%"')

        infos = [str(val.get(x)) for x in ['context', 'origin', 'fulltype']]
        infos = {y: x for x, y in zip(infos, ['Context', 'File Origin', 'Resource Type'])}
        infosTag = tag.td(infoBlock(infos), 'class="titlesub log" style="width:40%"')

        success = val['success']
        if success:
            getTag = tag.td('GET Success HTTP Code {}'.format(rcode), 'class="pass"')
        else:
            getTag = tag.td('GET Failure HTTP Code {}'.format(rcode), 'class="fail"')

        countsTag = tag.td(infoBlock(val['counts'], split='', ffunc=applyInfoSuccessColor), 'class="log"')

        rhead = ''.join([buttonTag, infosTag, getTag, countsTag])
        for x in [('tr',), ('table', 'class=titletable'), ('td', 'class=titlerow'), ('tr')]:
            rhead = wrapTag(''.join(rhead), *x)
        entry.append(rhead)

        # actual table

        rows = [list([str(vars(m)[x]) for x in LOG_ENTRY]) for m in val['messages'].values()]
        titles = ['Property Name', 'Value', 'Type', 'Exists', 'Result']
        widths = ['15', '30', '30', '10', '15']
        tableHeader = tableBlock(rows, titles, widths, ffunc=applySuccessColor)

        #    lets wrap table and errors and warns into one single column table
        tableHeader = tag.tr(tag.td((tableHeader)))

        infos_a = [str(val.get(x)) for x in ['uri'] if val.get(x) not in ['',None]]
        infos_a.append(rtime)

        # warns and errors
        errors = val['errors']
        if len(errors) == 0:
            errors = 'No errors'
        infos = errors.split('\n')
        errorTags = tag.tr(tag.td(infoBlock(infos), 'class="fail log"'))

        warns = val['warns']
        if len(warns) == 0:
            warns = 'No warns'
        infos = warns.split('\n')
        warnTags = tag.tr(tag.td(infoBlock(infos), 'class="warn log"'))

        payloadTag = tag.td(json.dumps(payload, indent=4, sort_keys=True), 'id=\'payload{}\' class=\'results log\''.format(cnt))

        tableHeader += errorTags
        tableHeader += warnTags
        tableHeader += payloadTag
        tableHeader = tag.table(tableHeader)
        tableHeader = tag.td(tableHeader, 'class="results" id=\'resNum{}\''.format(cnt))

        entry.append(tableHeader)

        # append
        htmlPage += ''.join([tag.tr(x) for x in entry])

    return wrapTag(wrapTag(htmlStrTop + wrapTag(htmlStrBodyHeader + htmlPage, 'table'), 'body'), 'html')


def writeHtml(string, path):
    with open(path, 'w', encoding='utf-8') as f:
        f.write(string)


def htmlLogScraper(htmlReport, output_name=None):
    outputLogName = os.path.split(htmlReport)[-1] if not output_name else output_name
    output = open('./{}.csv'.format(outputLogName),'w',newline='')
    csv_output = csv.writer(output)
    csv_output.writerow(['URI','Status','Response Time','Context','File Origin','Resource Type','Property Name','Value','Expected','Actual','Result'])
    htmlLog = open(htmlReport,'r')
    soup = BeautifulSoup(htmlLog, 'html.parser')
    glanceDetails = {}
    table = soup.find_all('table', {'class':'titletable'})
    for tbl in table:
        tr = tbl.find('tr')
        URIresp = tr.find('td',{'class':'title'}) # URI, response time, show results button
        URI = URIresp.text.partition('(')[0]
        responseTime = URIresp.text.partition('response time')[2].split(')')[0].strip(':s')
        StatusGET = tr.find('td',{'class':'pass'}) or tr.find('td',{'class':'fail'})
        if 'Success' in StatusGET.text:
            Status = '200'
        else:
            Status = '400'

        context,FileOrigin,ResourceType = ' ',' ',' '
        if 'Context:' in tr.find_all('td')[1].text:
            context = tr.find_all('td')[1].text.split('Context:')[1].split('File')[0]
        if 'File Origin'in tr.find_all('td')[1].text:
            FileOrigin = tr.find_all('td')[1].text.split('File Origin:')[1].split('Resource')[0]
        if 'Resource Type'in tr.find_all('td')[1].text:
            ResourceType = tr.find_all('td')[1].text.split('Resource Type:')[1]
        resNumHtml = tr.find('div', {'class':'button warn'})
        resNum = resNumHtml.attrs['onclick'].split(";")
        resNum = resNum[0].split("'")[1] if len(resNum) < 3 else resNum[1].split("'")[1]
        results = [ URI, Status, responseTime, context, FileOrigin, ResourceType ]
        glanceDetails[resNum] = results # mapping of results to their respective tables

    properties = soup.findAll('td',{'class':'results'})
    data = []
    for table in properties:
        tableID = table.attrs.get('id')
        if len(table.find_all('table')) == 0 or tableID in ['resMetadata', None]:
            continue
        tableBody = table.find_all('table')[-1]
        tableRows = tableBody.find_all('tr')[1:] #get rows from property tables excluding header
        for tr in tableRows:
            td = tr.find_all('td')
            row = [i.text for i in td]
            if tableID in glanceDetails:
                data.append(glanceDetails[tableID] + row)
    csv_output.writerows(data)
    output.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Get an excel sheet of details shown in the HTML reports for the Redfish Service Validator')
    parser.add_argument('htmllog' ,type=str, help = 'Path of the HTML log to be converted to csv format' )
    parser.add_argument('--dest' ,type=str, help = 'Name of output' )
    args = parser.parse_args()

    htmlLogScraper(args.htmllog, args.dest) 




from collections import Counter, OrderedDict
from redfish_service_validator.catalog import REDFISH_ABSENT, MissingSchemaError, ExcerptTypes, get_fuzzy_property, RedfishObject, RedfishType

from redfish_service_validator.helper import getNamespace, getNamespaceUnversioned, getType, checkPayloadConformance, stripCollection

import logging

my_logger = logging.getLogger()
my_logger.setLevel(logging.DEBUG)

def validateExcerpt(prop, val):
    # check Navprop if it's NEUTRAL or CONTAINS
    base = prop.Type.getBaseType()

    if base == 'entity':
        my_excerpt_type, my_excerpt_tags = prop.Type.excerptType, prop.Type.excerptTags

        my_new_type = stripCollection(prop.Type.fulltype)

        new_type_obj = prop.Type.catalog.getSchemaDocByClass(getNamespace(my_new_type)).getTypeInSchemaDoc(my_new_type)

        my_props = new_type_obj.createObject(prop.Name).populate(val).properties

        for name, innerprop in my_props.items():
            if not innerprop.HasSchema:
                continue
            valid_tagging = any([x in my_excerpt_tags for x in innerprop.Type.excerptTags]) or innerprop.Type.excerptTags == []
            if my_excerpt_type == ExcerptTypes.NEUTRAL:
                if innerprop.Type.excerptType == [ExcerptTypes.EXCLUSIVE] and innerprop.Exists:
                    my_logger.error('{}: Exclusive Excerpt {} should not exist in this Resource/ComplexType'.format(prop.Name, innerprop.Name))
                    return False
            if my_excerpt_type == ExcerptTypes.CONTAINS:
                if innerprop.Type.excerptType in [ExcerptTypes.ALLOWED, ExcerptTypes.EXCLUSIVE, ExcerptTypes.CONTAINS]:
                    if innerprop.Exists and not valid_tagging:
                        my_logger.error('{}: Excerpt tags of owner {} do not match property {} {}'.format(innerprop.Name, prop.Name, my_excerpt_tags, innerprop.Type.excerptTags))
                        return False
                elif innerprop.Exists:
                    my_logger.error('{}: Property is not a valid Excerpt'.format(innerprop.Name))
                    return False

    # check our prop if it's EXCLUSIVE
    if prop.Type.excerptType == ExcerptTypes.EXCLUSIVE and prop.Exists:
        my_logger.error('{}: Exclusive Excerpt should not exist in a Resource/Complex'.format(prop.Name))
        return False

    return True


def validateAction(act_fulltype, actionDecoded, all_actions):
    actionMessages, actionCounts = OrderedDict(), Counter()
    act_namespace, act_type = getNamespace(act_fulltype.strip('#')), getType(act_fulltype)
    actPass = False
    if act_type not in all_actions:
        my_logger.error('Action {} does not exist in Namespace {}'.format(act_type, act_namespace))
        actionCounts['errorActionBadName'] += 1
        actionMessages[act_fulltype] = (
                'Action', '-',
                'Yes' if actionDecoded != REDFISH_ABSENT else 'No',
                'FAIL')
    else:
        my_act = all_actions[act_type]
        actOptional = my_act.find('annotation', {'term': 'Redfish.Required'}) is not None
        if actionDecoded == REDFISH_ABSENT:
            if actOptional:
                actPass = True
            else:
                my_logger.error('{}: Mandatory action missing'.format(act_namespace))
                actionCounts['failMandatoryAction'] += 1
        if actionDecoded != REDFISH_ABSENT:
            # validate target
            target = actionDecoded.get('target')
            if target is None:
                my_logger.error('{}: target for action is missing'.format(act_namespace))
            elif not isinstance(target, str):
                my_logger.error('{}: target for action is malformed'.format(act_namespace))
                # check for unexpected properties
            for ap_name in actionDecoded:
                expected = ['target', 'title', '@Redfish.ActionInfo', '@Redfish.OperationApplyTimeSupport']
                if ap_name not in expected and '@Redfish.AllowableValues' not in ap_name:
                    my_logger.error('{}: Property "{}" is not allowed in actions property. \
                        Allowed properties are "{}", "{}", "{}", "{}" and "{}"'.format(act_namespace, ap_name, *expected, '*@Redfish.AllowableValues'))
            actPass = True
        if actOptional and actPass:
            actionCounts['optionalAction'] += 1
        elif actPass:
            actionCounts['passAction'] += 1
        else:
            actionCounts['failAction'] += 1
            
        actionMessages[act_fulltype] = (
                'Action', '-',
                'Yes' if actionDecoded != REDFISH_ABSENT else 'No',
                'Optional' if actOptional else 'PASS' if actPass else 'FAIL')
    return actionMessages, actionCounts


def validateEntity(service, prop, val, parentURI=""):
    """
    Validates an entity based on its uri given
    """
    name, val, autoExpand = prop.Name, val, prop.IsAutoExpanded
    excerptType = prop.Type.excerptType if prop.Type.Excerpt else ExcerptTypes.NEUTRAL
    my_logger.debug('validateEntity: name = {}'.format(name))

    # check for required @odata.id
    if not isinstance(val, dict):
        my_logger.info("{}: EntityType val is null/absent, not testing...".format(name))
        return False
    uri = val.get('@odata.id')
    if '@odata.id' not in val:
        if autoExpand: uri = parentURI + '#/{}'.format(name.replace('[', '/').strip(']'))
        else: uri = parentURI + '/{}'.format(name)
        if excerptType == ExcerptTypes.NEUTRAL:
            my_logger.error("{}: EntityType resource does not contain required @odata.id property, attempting default {}".format(name, uri))
            if parentURI == "":
                return False

    # check if the entity is truly what it's supposed to be
    # if not autoexpand, we must grab the resource
    if not autoExpand:
        success, data, response, delay = service.callResourceURI(uri)
        if success and response is not None:
            status = response.status
        else:
            status = -1
    else:
        success, data, response, delay = True, val, None, 0
        status = 200

    generics = ['Resource.ItemOrCollection', 'Resource.ResourceCollection', 'Resource.Item', 'Resource.Resource']
    my_type = prop.Type.fulltype
    if success and my_type in generics:
        return True
    elif success:
        my_target_type = data.get('@odata.type', 'Resource.Item').strip('#')
        # Attempt to grab an appropriate type to test against and its schema
        # Default lineup: payload type, collection type, property type
        my_type_chain = [str(x) for x in prop.Type.getTypeTree()]

        try:
            my_target_schema = prop.Type.catalog.getSchemaDocByClass(getNamespaceUnversioned(my_target_type))
        except MissingSchemaError:
            my_logger.error("{}: Could not get schema file for Entity check".format(name))
            return False

        if getNamespace(my_target_type) not in my_target_schema.classes:
            my_logger.error('{}: Linked resource reports version {} not in Schema'.format(name.split(':')[-1], my_target_type))
        else:
            my_target_type = my_target_schema.getTypeInSchemaDoc(my_target_type)
            all_target_types = [str(x) for x in my_target_type.getTypeTree()]
            if any(x in my_type_chain for x in all_target_types):
                return True
            else:
                my_logger.error('{}: Linked resource reports version {} not in Typechain' .format(name.split(':')[-1], my_target_type))
                return False
    else:
        if excerptType == ExcerptTypes.NEUTRAL:
            if "OriginOfCondition" in name:
                my_logger.verbose1("{}: GET of resource at URI {} returned HTTP {}, but was a temporary resource."
                                .format(name, uri, status if isinstance(status, int) and status >= 200 else "error"))
                return True

            else:
                my_logger.error("{}: GET of resource at URI {} returned HTTP {}. Check URI."
                                .format(name, uri, status if isinstance(status, int) and status >= 200 else "error"))
                return False
        else:
            return True
    return False


def validateComplex(service, sub_obj, prop_name, oem_check=True):
    subMsgs, subCounts = OrderedDict(), Counter()

    # Based on the object's properties, see if we need to insert a pattern to verify the contents
    # At this time, only the Identifier object has this type of check to ensure the DurableName matches the long description
    if "DurableName" in sub_obj.properties and "DurableNameFormat" in sub_obj.properties:
        if sub_obj.properties["DurableNameFormat"].Value == "NAA":
            sub_obj.properties["DurableName"].added_pattern = '^(([0-9A-Fa-f]{2}){8}){1,2}$'
        elif sub_obj.properties["DurableNameFormat"].Value == "FC_WWN":
            sub_obj.properties["DurableName"].added_pattern = '^([0-9A-Fa-f]{2}[:-]){7}([0-9A-Fa-f]{2})$'
        elif sub_obj.properties["DurableNameFormat"].Value == "UUID":
            sub_obj.properties["DurableName"].added_pattern = '([0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})'
        elif sub_obj.properties["DurableNameFormat"].Value == "EUI":
            sub_obj.properties["DurableName"].added_pattern = '^([0-9A-Fa-f]{2}[:-]){7}([0-9A-Fa-f]{2})$'
        elif sub_obj.properties["DurableNameFormat"].Value == "NGUID":
            sub_obj.properties["DurableName"].added_pattern = '^([0-9A-Fa-f]{2}){16}$'
        elif sub_obj.properties["DurableNameFormat"].Value == "MACAddress":
            sub_obj.properties["DurableName"].added_pattern = '^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$'

    for sub_name, sub_prop in sub_obj.properties.items():
        if not sub_prop.HasSchema and not sub_prop.Exists:
            subCounts['skipNoSchema'] += 1
            continue
        elif not sub_prop.HasSchema:
            my_logger.error('No Schema for sub_property {}'.format(sub_prop.Name))
            subCounts['errorNoSchema'] += 1
            continue
        new_msgs, new_counts = checkPropertyConformance(service, sub_name, sub_prop)
        subMsgs.update(new_msgs)
        subCounts.update(new_counts)

    jsonData = sub_obj.Value
    allowAdditional = sub_obj.Type.HasAdditional
    if prop_name != 'Actions':
        for key in [k for k in jsonData if k not in subMsgs and k not in sub_obj.properties and '@' not in k]:
            # note: extra subMsgs for "unchecked" properties
            item = jsonData.get(key)
            if not allowAdditional:
                my_logger.error('{} not defined in Complex {} {} (check version, spelling and casing)'
                                .format(key, prop_name, sub_obj.Type))
                subCounts['failAdditional.complex'] += 1
                subMsgs[key] = (displayValue(item), '-', '-', 'FAIL')
            else:
                my_logger.warning('{} not defined in schema Complex {} {} (check version, spelling and casing)'
                                .format(key, prop_name, sub_obj.Type))
                subCounts['unverifiedAdditional.complex'] += 1
                subMsgs[key] = (displayValue(item), '-', '-', 'Additional')
            
            fuzz = get_fuzzy_property(key, sub_obj.properties)
            if fuzz != key and fuzz in sub_obj.properties:
                subMsgs[fuzz] = ('-', '-', '-', 'INVALID')
                my_logger.error('Attempting {} (from {})?'.format(fuzz, key))
                my_new_obj = sub_obj.properties[fuzz].populate(item)
                new_msgs, new_counts = checkPropertyConformance(service, key, my_new_obj)
                subMsgs.update(new_msgs)
                subCounts.update(new_counts)
                subCounts['invalidNamedProperty.complex'] += 1

    successPayload, odataMessages = checkPayloadConformance(sub_obj.Value, '')

    if not successPayload:
        odataMessages['failPayloadError.complex'] += 1
        my_logger.error('{}: complex payload error, @odata property non-conformant'.format(str(sub_obj.Name)))

    if prop_name == 'Actions':
        actionMessages, actionCounts = OrderedDict(), Counter()

        # Get our actions from the object itself to test
        # Action Namespace.Type, Action Object
        my_actions = [(x.strip('#'), y) for x, y in sub_obj.Value.items() if x != 'Oem']
        if 'Oem' in sub_obj.Value.items():
            if oem_check:
                my_actions.extend([(x, y) for x, y in sub_obj.Value['Oem'].items()])
            else:
                actionCounts['oemActionSkip'] += len(sub_obj.Value['Oem'])

        # get ALL actions (but we don't need to test for them...)
        # ...
        # for new_act in sub_class.actions:
        #     new_act_name = '#{}.{}'.format(base_type, new_act)
        #     if new_act_name not in my_actions:
        #         my_actions.append((new_act_name, REDFISH_ABSENT))

        for act_name, actionDecoded in my_actions:
            if '@' in act_name:
                continue
            act_schema = sub_obj.Type.catalog.getSchemaDocByClass(getNamespace(act_name))
            act_class = act_schema.classes.get(getNamespace(act_name))

            a, c = validateAction(act_name, actionDecoded, act_class.actions)

            actionMessages.update(a)
            actionCounts.update(c)
        subMsgs.update(actionMessages)
        subCounts.update(actionCounts)
    return subMsgs, subCounts


def displayType(propTypeObject, is_collection=False):
    """
    Convert inputs propType and propRealType to a simple, human readable type
    :param propType: the 'Type' attribute from the PropItem.propDict
    :param propRealType: the 'realtype' entry from the PropItem.propDict
    :param is_collection: For collections: True if these types are for the collection; False if for a member
    :return: the simplified type to display
    """
    propRealType, propCollection = propTypeObject.getBaseType(), propTypeObject.IsCollection()
    propType = propTypeObject.fulltype
    # Edm.* and other explicit types
    if propRealType == 'Edm.Boolean' or propRealType == 'Boolean':
        disp_type = 'boolean'
    elif propRealType == 'Edm.String' or propRealType == 'String':
        disp_type = 'string'
    elif (propRealType.startswith('Edm.Int') or propRealType == 'Edm.Decimal' or
        propRealType == 'Edm.Double' or propRealType == 'Integer'):
        disp_type = 'number'
    elif propRealType == 'Edm.Guid':
        disp_type = 'GUID'
    elif propRealType == 'Edm.Primitive' or propRealType == 'Edm.PrimitiveType':
        disp_type = 'primitive'
    elif propRealType == 'Edm.DateTimeOffset':
        disp_type = 'date'
    elif propRealType == 'Edm.Duration':
        disp_type = 'duration'
    elif propRealType == 'Password':
        disp_type = 'password'
    elif propRealType == 'enum' or propRealType == 'deprecatedEnum' or propRealType == 'Enumeration':
        disp_type = 'string (enum)'
    elif propRealType.startswith('Edm.'):
        disp_type = propRealType.split('.', 1)[1]
    # Entity types
    elif propRealType == 'entity':
        member_type = getType(propType.replace('Collection(', '').replace(')', ''))
        if propCollection:
            if is_collection:
                disp_type = 'links: {}'.format(member_type)
            else:
                disp_type = 'link: {}'.format(member_type)
        else:
            disp_type = 'link to: {}'.format(member_type)
        if propTypeObject.AutoExpand:
            disp_type.replace('link', 'Expanded link')
    # Complex types
    elif propRealType == 'complex':
        if propCollection:
            member_type = getType(propType.replace('Collection(', '').replace(')', ''))
            if is_collection:
                disp_type = 'array of: {}'.format(member_type)
            else:
                disp_type = member_type
        else:
            disp_type = propType
    # Fallback cases
    elif len(propRealType) > 0:
        disp_type = propRealType
    elif len(propType) > 0:
        disp_type = propType
    else:
        disp_type = 'n/a'

    my_logger.debug('displayType: ({}, {}) -> {}'.format(propType, propRealType, disp_type))
    return disp_type


def displayValue(val, autoExpandName=None):
    """
    Convert input val to a simple, human readable value
    :param val: the value to be displayed
    :param autoExpand: optional, name of JSON Object if it is a referenceable member
    :return: the simplified value to display
    """
    if val is None:
        disp_val = '[null]'
    elif isinstance(val, dict) and len(val) == 1 and '@odata.id' in val:
        disp_val = 'Link: {}'.format(val.get('@odata.id'))
    elif isinstance(val, (int, float, str, bool)):
        disp_val = val
    elif autoExpandName is not None:
        disp_val = 'Referenceable object - see report {} listed below'.format(autoExpandName)
    else:
        disp_val = '[JSON Object]'

    my_logger.debug('displayValue: {} -> {}'.format(val, disp_val))
    return disp_val


def checkPropertyConformance(service, prop_name, prop, parent_name=None, parent_URI=""):
    """checkPropertyConformance

    Given a dictionary of properties, check the validitiy of each prop_name, and return a
    list of counted properties

    :param schemaObj:
    :param PropertyName:
    :param PropertyItem:
    :param decoded:
    :param ParentItem:
    :param parentURI:
    """

    resultList = OrderedDict()
    counts = Counter()

    my_logger.verbose1(prop_name)
    my_logger.verbose1("\tvalue: {} {}".format(prop.Value, type(prop.Value)))

    # Basic Validation of all properties
    prop_name = '.'.join([x for x in (parent_name, prop_name) if x])

    propNullable = prop.Type.IsNullable

    if not prop.SchemaExists:
        if not prop.Exists:
            my_logger.verbose1('{}: Item is skipped, no schema'.format(prop_name))
            counts['skipNoSchema'] += 1
            return {prop_name: ('-', '-', 'Yes' if prop.Exists else 'No', 'NoSchema')}, counts
        else:
            my_logger.error('{}: Item is present, but no schema found'.format(prop_name))
            counts['failNoSchema'] += 1
            return {prop_name: ('-', '-', 'Yes' if prop.Exists else 'No', 'FAIL')}, counts

    # check oem
    # rs-assertion: 7.4.7.2
    oem_check = service.config.get('oemcheck', True)
    if 'Oem' in prop_name and not oem_check:
        my_logger.verbose1('\tOem is skipped')
        counts['skipOem'] += 1
        return {prop_name: ('-', '-', 'Yes' if prop.Exists else 'No', 'OEM')}, counts

    # Parameter Passes
    paramPass = propMandatoryPass = propNullablePass = deprecatedPassOrSinceVersion = nullValid = True

    if prop.Type.IsMandatory:
        propMandatoryPass = True if prop.Exists else False
        my_logger.verbose1("\tMandatory Test: {}".format('OK' if propMandatoryPass else 'FAIL'))
    else:
        my_logger.verbose1("\tis Optional")
        if not prop.Exists:
            my_logger.verbose1("\tprop Does not exist, skip...")
            counts['skipOptional'] += 1
            return {prop_name: ( '-', displayType(prop.Type), 'Yes' if prop.Exists else 'No', 'Optional')}, counts

    # <Annotation Term="Redfish.Deprecated" String="This property has been Deprecated in favor of Thermal.v1_1_0.Thermal.Fan.Name"/>
    if prop.Type.Deprecated is not None and not prop.Type.IsMandatory:
        deprecatedPassOrSinceVersion = False
        counts['warnDeprecated'] += 1
        my_logger.warning('{}: The given property is deprecated: {}'.format(prop_name, prop.Type.Deprecated.get('String', '')))

    if prop.Type.Revisions is not None:
        for tag_item in prop.Type.Revisions:
            revision_tag = tag_item.find('PropertyValue', attrs={ 'EnumMember': 'Redfish.RevisionKind/Deprecated', 'Property': 'Kind'})
            if revision_tag and not prop.Type.IsMandatory:
                desc_tag = tag_item.find('PropertyValue', attrs={'Property': 'Description'})
                version_tag = tag_item.find('PropertyValue', attrs={'Property': 'Version'})
                deprecatedPassOrSinceVersion = version_tag.attrs.get('String', False) if version_tag else False
                counts['warnDeprecated'] += 1
                if desc_tag:
                    my_logger.warning('{}: The given property is deprecated: {}'.format(prop_name, desc_tag.attrs.get('String', '')))
                else:
                    my_logger.warning('{}: The given property is deprecated'.format(prop_name))

    # Note: consider http://docs.oasis-open.org/odata/odata-csdl-xml/v4.01/csprd01/odata-csdl-xml-v4.01-csprd01.html#_Toc472333112
    # Note: make sure it checks each one
    # propCollectionType = PropertyDict.get('isCollection')
    propRealType, isCollection = prop.Type.getBaseType(), prop.Type.IsCollection()

    excerptPass = True
    if not isCollection and isinstance(prop.Value, list): 
        my_logger.error('{}: Value of property is an array but is not a Collection'.format(prop_name))
        counts['failInvalidArray'] += 1
        return {prop_name: ( '-', displayType(prop.Type, is_collection=True), 'Yes' if prop.Exists else 'No', 'FAIL')}, counts

    if isCollection and prop.Value is None:
        # illegal for a collection to be null
        if 'EventDestination.v1_0_0.HttpHeaderProperty' == str(prop.Type.fulltype):
            # HttpHeaders in EventDestination has non-conformant details in the long description we need to allow to not break existing implementations
            my_logger.info('Value HttpHeaders can be Null')
            propNullable = True
            resultList[prop_name] = ('Array (size: null)', displayType(prop.Type, is_collection=True), 'Yes' if prop.Exists else 'No', '...')
        else:
            my_logger.error('{}: Value of Collection property is null but Collections cannot be null, only their entries'.format(prop_name))
            counts['failNullCollection'] += 1
            return {prop_name: ( '-', displayType(prop.Type, is_collection=True), 'Yes' if prop.Exists else 'No', 'FAIL')}, counts
    elif isCollection and prop.Value is not None:
        # note: handle collections correctly, this needs a nicer printout
        # rs-assumption: do not assume URIs for collections
        # rs-assumption: check @odata.count property
        # rs-assumption: check @odata.link property
        my_logger.verbose1("\tis Collection")
        if prop.Value == REDFISH_ABSENT:
            resultList[prop_name] = ('Array (absent) {}'.format(len(prop.Value)),
                                displayType(prop.Type, is_collection=True),
                                'Yes' if prop.Exists else 'No', 'PASS' if propMandatoryPass else 'FAIL')
        elif not isinstance(prop.Value, list):
            my_logger.error('{}: property is expected to contain an array'.format(prop_name))
            counts['failInvalidArray'] += 1
            resultList[prop_name] = ('-', displayType(prop.Type, is_collection=True), 'Yes' if prop.Exists else 'No', 'FAIL')
            return resultList, counts
        else:
            resultList[prop_name] = ('Array (size: {})'.format(len(prop.Value)), displayType(prop.Type, is_collection=True), 'Yes' if prop.Exists else 'No', '...')

    # If we're validating a complex object
    if propRealType == 'complex':
        result_str = 'complex'
        if prop.Type.IsMandatory and not prop.Exists:
            my_logger.error("{}: Mandatory prop does not exist".format(prop_name))
            counts['failMandatoryExist'] += 1
            result_str = 'FAIL'

        if not prop.Exists:
            return resultList, counts

        if prop.IsCollection:
            resultList[prop_name] = ('Array (size: {})'.format(len(prop.Value)), displayType(prop.Type, is_collection=True), 'Yes' if prop.Exists else 'No', result_str)
            object_list = prop.Value
        else:
            resultList[prop_name] = ('[JSON Object]', displayType(prop.Type), 'Yes' if prop.Exists else 'No', result_str)
            object_list = [prop]
    
        for n, sub_obj in enumerate(object_list):
            try:
                if sub_obj.Value is None:
                    if prop.Type.IsNullable or 'EventDestination.v1_0_0.HttpHeaderProperty' == str(prop.Type.fulltype):
                        # HttpHeaders in EventDestination has non-conformant details in the long description we need to allow to not break existing implementations
                        counts['pass'] += 1
                        result_str = 'PASS'
                    else:
                        my_logger.error('{}: Property is null but is not Nullable'.format(prop_name))
                        counts['failNullable'] += 1
                        result_str = 'FAIL'
                    if isinstance(prop, RedfishObject):
                        resultList['{}.[Value]'.format(prop_name)] = ('[null]', displayType(prop.Type),
                                                                        'Yes' if prop.Exists else 'No', result_str)
                    else:
                        resultList['{}.[Value]#{}'.format(prop_name, n)] = ('[null]', displayType(prop.Type), 'Yes' if prop.Exists else 'No', result_str)
                else:
                    subMsgs, subCounts = validateComplex(service, sub_obj, prop_name, oem_check)
                    if isCollection:
                        subMsgs = {'{}[{}].{}'.format(prop_name, n, x): y for x, y in subMsgs.items()}
                    elif isinstance(prop, RedfishObject):
                        subMsgs = {'{}.{}'.format(prop_name, x): y for x, y in subMsgs.items()}
                    else:
                        subMsgs = {'{}.{}#{}'.format(prop_name, x, n): y for x, y in subMsgs.items()}
                    resultList.update(subMsgs)
                    counts.update(subCounts)
            except Exception as ex:
                my_logger.verbose1('Exception caught while validating Complex', exc_info=1)
                my_logger.error('{}: Could not finish check on this property ({})'.format(prop_name, str(ex)))
                counts['exceptionPropCheck'] += 1
        return resultList, counts

    # Everything else...
    else:
        propValueList = prop.Value if prop.IsCollection else [prop.Value]
        for cnt, val in enumerate(propValueList):
            appendStr = (('[' + str(cnt) + ']') if prop.IsCollection else '')
            sub_item = prop_name + appendStr

            excerptPass = validateExcerpt(prop, val)

            if isinstance(val, str):
                if val == '' and prop.Type.Permissions == 'OData.Permission/Read':
                    my_logger.warning('{}: Empty string found - Services should omit properties if not supported'.format(sub_item))
                    nullValid = False
                if val.lower() == 'null':
                    my_logger.warning('{}: "null" string found - Did you mean to use an actual null value?'.format(sub_item))
                    nullValid = False

            if prop.Exists:
                paramPass = propNullablePass = True
                if val is None:
                    if propNullable:
                        my_logger.debug('Property {} is nullable and is null, so Nullable checking passes'.format(sub_item))
                    else:
                        propNullablePass = False
                
                if isinstance(prop.Type, str) and 'Edm.' in prop.Type:
                    try:
                        paramPass = prop.Exists and prop.validate_basic(val, prop.Type)
                    except ValueError as e:
                        my_logger.error('{}: {}'.format(prop.Name, e))  # log this
                        paramPass = False
                elif isinstance(prop.Type, RedfishType):
                    try:
                        paramPass = prop.Type.validate(val, prop.added_pattern)
                    except ValueError as e:
                        my_logger.error('{}: {}'.format(prop.Name, e))  # log this
                        paramPass = False

                if propRealType == 'entity':
                    paramPass = validateEntity(service, prop, val)
                


            # Render our result
            my_type = prop.Type.fulltype

            if all([paramPass, propMandatoryPass, propNullablePass, excerptPass]):
                my_logger.verbose1("\tSuccess")
                counts['pass'] += 1
                result_str = 'PASS'
                if deprecatedPassOrSinceVersion is False:
                    result_str = 'Deprecated'
                if isinstance(deprecatedPassOrSinceVersion, str):
                    result_str = 'Deprecated/{}'.format(deprecatedPassOrSinceVersion)
                if not nullValid:
                    counts['invalidPropertyValue'] += 1
                    result_str = 'WARN'
            else:
                my_logger.verbose1("\tFAIL")
                counts['err.' + str(my_type)] += 1
                result_str = 'FAIL'
                if not paramPass:
                    if prop.Type.IsMandatory:
                        counts['failMandatoryProp'] += 1
                    else:
                        counts['failProp'] += 1
                elif not propMandatoryPass:
                    my_logger.error("{}: Mandatory prop does not exist".format(sub_item))
                    counts['failMandatoryExist'] += 1
                elif not propNullablePass:
                    my_logger.error('{}: Property is null but is not Nullable'.format(sub_item))
                    counts['failNullable'] += 1
                elif not excerptPass:
                    counts['errorExcerpt'] += 1
                    result_str = 'errorExcerpt'

            resultList[sub_item] = (
                    displayValue(val, sub_item if prop.IsAutoExpanded else None), displayType(prop.Type),
                    'Yes' if prop.Exists else 'No', result_str)

        return resultList, counts